# Tests of in-memory notebooks
#
# Copyright (C) 2016--2020 Simon Dobson
# 
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import *

import unittest
from datetime import datetime

    
class LabNotebookTests(unittest.TestCase):

    def setUp(self):
        '''Set up a notebook.'''
        self._nb = LabNotebook()

    def _resultsdict(self):
        '''Set up a results dict populated with plausible metadata.'''
        _rc = Experiment.resultsdict()
        _rc[Experiment.METADATA][Experiment.EXPERIMENT] = str(self.__class__)
        _rc[Experiment.METADATA][Experiment.START_TIME] = datetime.now()
        _rc[Experiment.METADATA][Experiment.END_TIME] = datetime.now()
        _rc[Experiment.METADATA][Experiment.SETUP_TIME] = 10
        _rc[Experiment.METADATA][Experiment.EXPERIMENT_TIME] = 20
        _rc[Experiment.METADATA][Experiment.TEARDOWN_TIME] = 10
        _rc[Experiment.METADATA][Experiment.ELAPSED_TIME] = 40
        _rc[Experiment.METADATA][Experiment.STATUS] = True
        return _rc

    def testEmptyNotebook( self ):
        '''Test creating an empty notebook'''
        nb = LabNotebook("test")
        self.assertEqual(nb.name(), "test")
        self.assertFalse(nb.isPersistent())
        
    def testOneResultSet(self):
        '''Test we can ignore result sets if we want to.'''
        self.assertIsNotNone(self._nb.current())
        self.assertCountEqual(self._nb.resultSets(), [ LabNotebook.DEFAULT_RESULTSET ])

    def testAddingResultSets(self):
        '''Test we can add result sets .'''
        self._nb.addResultSet('second')
        self.assertCountEqual(self._nb.resultSets(), [ LabNotebook.DEFAULT_RESULTSET, 'second' ])
        self.assertEqual(self._nb.numberOfResultSets(), 2)
        self.assertEqual(len(self._nb), 2)

    def _resultsEqual(self, df, rc):
        '''Check that the dataframe contains a row with the given results.

        :param df: the dataframe
        :param rc: the results dict
        :returns: True if there's a corresponding row'''
        for d in [ Experiment.PARAMETERS, Experiment.RESULTS ]:  # ignore metadata
            for k in rc[d].keys():
                df = df[df[k] == rc[d][k]]
        return (len(df) > 0)

    def testDifferentParameters(self):
        '''Test different rule sets maintain different parameter sets.'''
        rc1 = self._resultsdict()
        rc1[Experiment.PARAMETERS]['a'] = 1
        rc1[Experiment.PARAMETERS]['b'] = 2
        rc1[Experiment.RESULTS]['first'] = 3
        self._nb.addResult(rc1)

        self._nb.addResultSet('other')
        rc2 = self._resultsdict()
        rc2[Experiment.PARAMETERS]['c'] = 6
        rc2[Experiment.PARAMETERS]['b'] = 1
        rc2[Experiment.RESULTS]['second'] = 12
        self._nb.addResult(rc2)

        self._nb.select(LabNotebook.DEFAULT_RESULTSET)
        self.assertTrue(self._resultsEqual(self._nb.current().dataframeFor(rc1[Experiment.PARAMETERS]), rc1))
        with self.assertRaises(Exception):
            self._nb.resultsFor(rc2[Experiment.PARAMETERS])

        self._nb.select('other')
        self.assertTrue(self._resultsEqual(self._nb.current().dataframeFor(rc2[Experiment.PARAMETERS]), rc2))
        with self.assertRaises(Exception):
            self._nb.resultsFor(rc1[Experiment.PARAMETERS])
        
    def testPendingResultsAreNotified(self):
        '''Test the notebook records pending results correctly.'''
        rc1 = self._resultsdict()
        rc1[Experiment.PARAMETERS]['a'] = 1
        rc1[Experiment.PARAMETERS]['b'] = 2
        rc1[Experiment.RESULTS]['first'] = 3

        self._nb.addPendingResult(rc1[Experiment.PARAMETERS], '1234')
        self.assertFalse(self._nb.ready())

        self._nb.resolvePendingResult(rc1, '1234')
        with self.assertRaises(Exception):
            self._nb.resolvePendingResult(rc1, '1234')
        self.assertTrue(self._nb.ready())

        self._nb.addPendingResult(rc1[Experiment.PARAMETERS], '4567')
        self._nb.cancelPendingResult('4567')
        with self.assertRaises(Exception):
            self._nb.resolvePendingResult(rc1, '4567')
        self.assertTrue(self._nb.ready())

    def testDataframe(self):
        '''Test dataframe access gets all results and respects success flag.'''
        rc1 = self._resultsdict()
        rc1[Experiment.METADATA][Experiment.STATUS] = True
        rc1[Experiment.PARAMETERS]['a'] = 1
        rc1[Experiment.PARAMETERS]['b'] = 2
        rc1[Experiment.RESULTS]['first'] = 3
        self._nb.addResult(rc1)
        rc1[Experiment.PARAMETERS]['b'] = 6
        rc1[Experiment.RESULTS]['first'] = 12
        self._nb.addResult(rc1)
        rc1[Experiment.METADATA][Experiment.STATUS] = False
        rc1[Experiment.PARAMETERS]['b'] = 8
        del rc1[Experiment.RESULTS]['first']
        self._nb.addResult(rc1)

        self.assertEqual(self._nb.numberOfResults(), 3)
        self.assertEqual(len(self._nb.results()), 3)
        self.assertEqual(len(self._nb.dataframe(only_successful=False)), 3)
        self.assertEqual(len(self._nb.dataframe()), 2)

    def testResultFraction(self):
        '''Test we compute the result fractions properly.'''
        rc1 = self._resultsdict()
        rc1[Experiment.PARAMETERS]['a'] = 10
        rc2 = self._resultsdict()
        rc2[Experiment.PARAMETERS]['a'] = 20
        rc3 = self._resultsdict()
        rc3[Experiment.PARAMETERS]['a'] = 30

        # check with only results
        self._nb.addResult(rc1)
        self.assertEqual(self._nb.readyFraction(), 1.0)

        # check with a pending result as well
        self._nb.addPendingResult(rc2[Experiment.PARAMETERS], '1234')
        self.assertEqual(self._nb.readyFraction(), 0.5)

        # check per-result set behaviour
        self._nb.addResultSet('second')
        self.assertEqual(self._nb.readyFraction(), 1.0)
        self._nb.addResult(rc3)
        self._nb.select(LabNotebook.DEFAULT_RESULTSET)
        self.assertEqual(self._nb.readyFraction(), 0.5)

    def testPendingResultsByTag(self):
        '''Test we can identify a result set by tag when querying.'''
        rc1 = self._resultsdict()
        rc1[Experiment.PARAMETERS]['a'] = 10
        rc2 = self._resultsdict()
        rc2[Experiment.PARAMETERS]['a'] = 20
        rc3 = self._resultsdict()
        rc3[Experiment.PARAMETERS]['a'] = 30

        # first result set
        self._nb.addResultSet('first')
        self._nb.addResult(rc1)
        self._nb.addResult(rc2)
        self._nb.addPendingResult(rc2[Experiment.PARAMETERS], '1234')

        # second
        self._nb.addResultSet('second')
        self._nb.addResult(rc3)

        # test default and tagged behaviour
        self.assertEqual(self._nb.numberOfResults(), 1)
        self.assertEqual(self._nb.numberOfPendingResults(), 0)
        self.assertEqual(self._nb.numberOfResults('first'), 2)
        self.assertEqual(self._nb.numberOfPendingResults('first'), 1)
        self.assertEqual(self._nb.numberOfResults('second'), 1)
        self.assertEqual(self._nb.numberOfPendingResults('second'), 0)
        self.assertEqual(len(self._nb.pendingResults()), 0)
        self.assertEqual(len(self._nb.pendingResults('first')), 1)

    def testAllPendingResults(self):
        '''Test we can retrieve all the penging results from across a notebook.'''
        rc1 = self._resultsdict()
        rc1[Experiment.PARAMETERS]['a'] = 10
        rc2 = self._resultsdict()
        rc2[Experiment.PARAMETERS]['a'] = 20
        rc3 = self._resultsdict()
        rc3[Experiment.PARAMETERS]['a'] = 30
        rc4 = self._resultsdict()
        rc4[Experiment.PARAMETERS]['a'] = 40

        # first result set
        self._nb.addResultSet('first')
        self._nb.addResult(rc1)
        self._nb.addResult(rc2)
        self._nb.addPendingResult(rc2[Experiment.PARAMETERS], '1234')

        # second
        self._nb.addResultSet('second')
        self._nb.addResult(rc3)
        self._nb.addPendingResult(rc4[Experiment.PARAMETERS], '2345')

        # check we get them all
        self.assertCountEqual(self._nb.allPendingResults(), ['1234', '2345'])
        self.assertEqual(self._nb.numberOfAllPendingResults(), 2)

    def testTaggedResultsAndDataframes(self):
        '''Test we can retrieve correctly from different result sets.'''
        rc1 = self._resultsdict()
        rc1[Experiment.PARAMETERS]['a'] = 10
        rc2 = self._resultsdict()
        rc2[Experiment.PARAMETERS]['a'] = 20
        rc3 = self._resultsdict()
        rc3[Experiment.PARAMETERS]['a'] = 30
        rc4 = self._resultsdict()
        rc4[Experiment.PARAMETERS]['a'] = 40

        # first result set
        self._nb.addResultSet('first')
        self._nb.addResult(rc1)
        self._nb.addResult(rc2)
        self._nb.addPendingResult(rc2[Experiment.PARAMETERS], '1234')

        # second
        self._nb.addResultSet('second')
        self._nb.addResult(rc3)
        self._nb.addPendingResult(rc4[Experiment.PARAMETERS], '2345')

        # check retrieval
        self._nb.select(LabNotebook.DEFAULT_RESULTSET)
        self.assertEqual(len(self._nb.dataframe()), 0)
        self.assertEqual(len(self._nb.dataframe(tag=LabNotebook.DEFAULT_RESULTSET)), 0)
        self.assertEqual(len(self._nb.dataframe(tag='first')), 2)
        self.assertEqual(len(self._nb.dataframe(tag='second')), 1)
        self.assertEqual(len(self._nb.dataframeFor(dict(a=10), tag='first')), 1)
        self.assertEqual(len(self._nb.dataframeFor(dict(a=40), tag='first')), 0)
        self.assertEqual(len(self._nb.results(tag='first')), 2)
        self.assertEqual(len(self._nb.results(tag='second')), 1)
        self.assertEqual(len(self._nb.resultsFor(dict(a=20), tag='first')), 1)
        self.assertEqual(len(self._nb.resultsFor(dict(a=30), tag='first')), 0)

    def testAddList(self):
        '''Test we can add a list of results dicts.'''
        rc1 = self._resultsdict()
        rc1[Experiment.PARAMETERS]['a'] = 10
        rc2 = self._resultsdict()
        rc2[Experiment.PARAMETERS]['a'] = 20
        rc3 = self._resultsdict()
        rc3[Experiment.PARAMETERS]['a'] = 30
        self._nb.addResult([rc1, rc2, rc3])
        self.assertEqual(self._nb.numberOfResults(), 3)
        df = self._nb.dataframe()
        vals = df['a']
        self.assertCountEqual(vals, [10, 20, 30])

    def testAddNested(self):
        '''Test we can add a set of nested results, as we get from a repeated experiment.'''
        rc = self._resultsdict()
        rc1 = self._resultsdict()
        rc1[Experiment.PARAMETERS]['a'] = 10
        rc2 = self._resultsdict()
        rc2[Experiment.PARAMETERS]['a'] = 20
        rc3 = self._resultsdict()
        rc3[Experiment.PARAMETERS]['a'] = 30

        # construct the nested experiment
        rc[Experiment.RESULTS] = [rc1, rc2, rc3]
        self._nb.addResult([rc1, rc2, rc3])

        self.assertEqual(self._nb.numberOfResults(), 3)
        df = self._nb.dataframe()
        vals = df['a']
        self.assertCountEqual(vals, [10, 20, 30])

    def testFinish(self):
        '''Test we can finish (lock) an entire notebook.'''
        rc1 = self._resultsdict()
        rc1[Experiment.PARAMETERS]['a'] = 10
        rc2 = self._resultsdict()
        rc2[Experiment.PARAMETERS]['a'] = 20
        rc3 = self._resultsdict()
        rc3[Experiment.PARAMETERS]['a'] = 30
        rc4 = self._resultsdict()
        rc4[Experiment.PARAMETERS]['a'] = 40

        # first result set
        self._nb.addResultSet('first')
        self._nb.addResult(rc1)
        self._nb.addResult(rc2)

        # second
        self._nb.addResultSet('second')
        self._nb.addResult(rc3)
        self._nb.addPendingResult(rc4[Experiment.PARAMETERS], '2345')

        # lock the notebook
        self._nb.finish()

        # check we can't add new result sets
        with self.assertRaises(LabNotebookLockedException):
            self._nb.addResultSet('third')

        # check the result sets are locked and finished correctly, with
        # any pending results cancelled
        rs = self._nb.select('first')
        self.assertTrue(rs.isLocked())
        self.assertEqual(rs.numberOfResults(), 2)
        rs = self._nb.select('second')
        self.assertTrue(rs.isLocked())
        self.assertEqual(rs.numberOfPendingResults(), 0)
        self.assertEqual(rs.numberOfResults(), 2)
        rcs = self._nb.resultsFor(rc4[Experiment.PARAMETERS])
        self.assertEqual(len(rcs), 1)
        self.assertFalse(rcs[0][Experiment.METADATA][Experiment.STATUS])

# TODO: Test we can add metadata

if __name__ == '__main__':
    unittest.main()

 


# Simulation "lab" experiment management, sequential version
#
# Copyright (C) 2016--2020 Simon Dobson
# 
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import LabNotebook, Experiment, ResultsDict
from pandas import DataFrame                                   # type: ignore
from typing import Dict, List, Any


class Lab(object):
    """A laboratory for computational experiments.

    A :class:`Lab` conducts an experiment at different points in a
    multi-dimensional parameter space.  The default performs all the
    experiments locally; sub-classes exist to perform remote parallel
    experiments.

    A :class:`Lab` stores its result in a notebook, an instance of :class:`LabNotebook`.
    By default the base :class:`Lab` class uses an in-memory notebook, essentially
    just a dict; sub-classes use persistent notebooks to manage larger
    sets of experiments.
    """

    def __init__( self, notebook : LabNotebook = None ):
        """Create an empty lab.

        :param notebook: the notebook used to store results (defaults to an empty :class:`LabNotebook`)"""
        if notebook is None:
            self._notebook = LabNotebook()
        else:
            self._notebook = notebook
        self._parameters : Dict[str, Any] = dict()

    def notebook(self) -> LabNotebook:
        """Return the notebook being used by this lab.

        :returns: the notebook"""
        return self._notebook


    # ---------- Protocol ----------

    def open( self ):
        """Open a lab for business. Sub-classes might insist the they are
        opened and closed explicitly when experiments are being performed.
        The default does nothing."""
        pass

    def close( self ):
        """Shut down a lab. Sub-classes might insist the they are
        opened and closed explicitly when experiments are being performed.
        The default does nothing."""
        pass

    def updateResults( self ):
        """Update the lab's results. This method is called by all other methods
        that return results in some sense, and may be overridden to let the results
        "catch up" with external processing. The default does nothing."""
        pass

    def recreate(self):
        '''Return a structure describing this lab in enough detail to reconstruct it,
        consisting of the name of the class and a dict of any arguments that it needs.
        Sub-classes should call the base method to fill in any defaults and then add
        any arguments they need to the dict.

        :returns: a (classname, args) pair
        '''
        n = '{modulename}.{classname}'.format(modulename = self.__class__.__module__,
                                              classname = self.__class__.__name__)
        args = dict()
        return (n, args)


    # ---------- Managing experimental parameters ----------

    def addParameter(self, k : str, r : Any):
        """Add a parameter to the experiment's parameter space. k is the
        parameter name, and r is its range. The range can be a single value
        or a list, or any other iterable. (Strings are counted as single values.)

        :param k: parameter name
        :param r: parameter range"""
        if isinstance(r, str):
            # strings are single values
            self._parameters[k] = [ r ]
        else:
            try:
                # try to unpack using iterator
                self._parameters[k] = list(r)
            except TypeError:
                # not iterable, a single value
                self._parameters[k] = [ r ]

    def parameters(self) -> List[str]:
        """Return a list of parameter names.

        :returns: a list of parameter names"""
        return list(self._parameters.keys())

    def __len__(self) -> int:
        """The length of an experiment is the total number of data points
        that will be explored.

        :returns: the length of the experiment"""
        n = 1
        for p in self.parameters():
            n = n * len(self._parameters[p])
        return n
        
    def __getitem__(self, k : str) -> Any:
        """Access a parameter range using array notation.

        :param k: parameter name
        :returns: the parameter range"""
        return self._parameters[k]

    def __setitem__(self, k : str, r : Any) -> Any:
        """Add a parameter using array notation.

        :param k: the parameter name
        :param r: the parameter range"""
        return self.addParameter(k, r)

    def _crossProduct(self, ls : List[str]) -> List[Dict[str, Any]]:
        """Internal method to generate the cross product of all parameter
        values, creating the parameter space for the experiment.

        :param ls: an array of parameter names
        :returns: list of dicts"""
        p = ls[0]
        ds = []
        if len(ls) == 1:
            # last parameter, convert range to a dict
            for i in self._parameters[p]:
                dp = dict()
                dp[p] = i
                ds.append(dp)
        else:
            # for other parameters, create a dict combining each value in
            # the range to a copy of the dict of other parameters
            ps = self._crossProduct(ls[1:])
            for i in self._parameters[p]:
                for d in ps:
                    dp = d.copy()
                    dp[p] = i
                    ds.append(dp)

        # return the complete parameter space
        return ds
           
    def parameterSpace(self) -> List[Dict[str, Any]]:
        """Return the parameter space of the experiment as a list of dicts,
        with each dict mapping each parameter name to a value.

        :returns: the parameter space as a list of dicts"""
        ps = self.parameters()
        if len(ps) == 0:
            return []
        else:
            return self._crossProduct(ps)


    # ---------- Running experiments ----------

    def runExperiment(self, e : Experiment):
        """Run an experiment over all the points in the parameter space.
        The results will be stored in the notebook.

        :param e: the experiment"""

        # create the parameter space
        ps = self.parameterSpace()

        # run the experiment at each point
        nb = self.notebook()
        for p in ps:
            #print "Running {p}".format(p = p)
            res = e.set(p).run()
            nb.addResult(res)

        # commit the results
        nb.commit()


    # ---------- Accessing results ----------

    def dataframe(self, only_successful : bool =True) -> DataFrame:
        """Return the current results as a pandas DataFrame after resolving
        any pending results that have completed. This makes
        use of the underlying notebook's current result set. For finer control,
        access the notebook's :meth:`LabNotebook.dataframe` or
        :meth:LabNotebook.dataframeFor` methods directly. 

        :param only_successful: only return successful results
        :returns: the resulting dataset as a DataFrame"""
        self.updateResults()
        return self._notebook.dataframe(only_successful=only_successful)

    def results(self) -> List[ResultsDict]:
        '''Return the current results as a list of results dicts after resolving
        any pending results that have completed. This makes
        use of the underlying notebook's current result set. For finer control,
        access the notebook's :meth:`LabNotebook.results` or
        :meth:LabNotebook.resultsFor` methods directly.
        
        Note that this approach to acquiring results is a lot slower and more
        memory-hungry than using :meth:`dataframe`, but may be useful for small
        sets of results that benefit from a more Pythonic intertface.'''
        self.updateResults()
        return self._notebook.results()

    def ready(self, tag : str =None) -> bool:
        """Test whether all the results are ready in the tagged result
        set -- that is, none are pending.

        :param tag: (optional) the result set to check (default is the current result set)
        :returns: True if the results are in"""
        self.updateResults()
        return self._notebook.ready(tag)

    def readyFraction(self, tag : str =None) -> float:
        '''Return the fraction of results available (not pending) in the
        tagged result set after first updating the results.

        :param tag: (optional) the result set to check (default is the current result set)
        :returns: the ready fraction'''
        self.updateResults()
        return self._notebook.readyFraction(tag)




# Simulation "lab notebook" for collecting results, JSON version
#
# Copyright (C) 2016--2020 Simon Dobson
# 
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import LabNotebook, Experiment, PackageContactInfo
import os
import sys
import json
import re
import numpy
from datetime import datetime
import dateutil.parser
from typing import Any, Dict


class MetadataEncoder(json.JSONEncoder):
    """Add support for encoding Python datetime objects within
    JSON objects, using the standard ISO string representation.
    (Plus a few other minor tweaks to get things to work more smoothly.)"""

    def default( self, o : Any) -> Any:
        """If o is a datetime object, convert it to an ISO string. If it is an
        exception, convert it to a string. If it is a numpy int, coerce it to
        a Python int.

        :param o: the field to serialise
        :returns: a string encoding of the field"""
        if isinstance(o, datetime):
            # date/time, return an ISO formatted string
            return o.isoformat()
        elif isinstance(o, Exception):
            # exception, stringify it
            return str(o)
        elif isinstance(o, numpy.integer):
            # numpy ints inherit differently between Python 2 and 3
            # see https://bugs.python.org/issue24313
            return int(o)
        else:
            # everything else, pass through
            return super(MetadataEncoder, self).default(o)


class JSONLabNotebook(LabNotebook):
    '''A lab notebook that persists intself to a JSON file. This is
    the most basic kind of persistent notebook, readable by
    virtually any tooling.

    Using JSON presents some disadvantages, as not all types can be
    represented. Specifically, exceptions from the metadata of failed
    experiments (with :attr:`Experiment.EXCEPTION`) will be saved as strings.
    We also need to convert `datetime` objects to ISO-format strings
    when saving.

    :param name: JSON file to persist the notebook to
    :param create: if True, erase existing file (defaults to False)
    :param description: free text description of the notebook
    '''

    # Structure of the file
    VERSION = 'version'                              #: Tag for version number (missing for v1).

    def __init__(self, name : str, create : bool =False, description : str =None):
        super(JSONLabNotebook, self).__init__(name, description)

        # check for the file already existing
        if os.path.isfile(name):
            # file exists, do we load it or create into it?
            if create:
                self._create(name)
            else:
                self._load(name)

                # use the description we were given, if there is one
                if description is not None:
                    self.setDescription(description)

    def isPersistent(self) -> bool:
        """Return True to indicate the notebook is persisted to a JSON file.

        :returns: True"""
        return True

    def commit(self):
        """Persist to disc."""
        self._save(self.name())
        
    def _create(self, fn : str):
        '''Create an empty JSON file for this notebook.

        :param fn: the file name'''
        with open(fn, 'w') as f:
            f.write('')

    def _load(self, fn : str):
        """Retrieve the notebook from the given file.

        :param fn: the file name"""
        if os.path.getsize(fn) > 0:
            with open(fn, "r") as f:
                # load the JSON object
                s = f.read()
                j = json.loads(s)

                # check version
                if self.VERSION in j:
                    # we have a version string, check it's ours
                    v = j[self.VERSION] 
                    if v == '2':
                        self._newload(j)
                    else:
                        raise Exception('Unhandled JSON file format {v}'.format(v=v))
                else:
                    # no version string, do the old-style load
                    print('Version 1 JSON format, notebook may have import errors', file=sys.stderr)
                    self._oldload(j)

    def _oldload(self, j : Dict[str, Any]):
        '''Load an old-format file.

        In this format, all results were held in dicts keyed by a key synthesised from the
        parameter names and values. Pending results were held as a mapping from job ids
        to these synthetic keys.

        :param j: the old-style JSON object'''

        # description
        self.setDescription(j['description'])

        # results
        # held as mapping from key to list of results dicts for same parameters
        for k in j['results]']:
            rcs = j['results'][k]
            for rc in rcs:
                meta = rc[Experiment.METADATA]
                for k in meta:
                    if k in [ Experiment.START_TIME, Experiment.END_TIME ]:
                        meta[k] = dateutil.parser.parse(meta[k])    # patch ISO-format strings to datetime objects
                self.addResult(rc)

        # pending results
        # held as mapping from job id to key, which need to be unpacked to retrieve the actual parameters
        # key held as string containing a sequence of of "<name>:[[<value>]];" for each parameter
        # sd: which idiot decided '[[...]]' was good bracketing???...
        pattern = re.compile('([a-zA-Z0-9\-\.]+)=\[\[([^\]]+)\]\]')
        pendings = dict(j['pending'])
        for jobid in pendings:
            key = pendings[jobid]

            # convert key to params
            params = dict()
            ps = key.split(';')
            for p in ps[0:-1]:      # terminating ; needs to be ignored
                m = pattern.match(p)
                n = m.group(1)
                v = m.group(2)
                if n in [ Experiment.START_TIME, Experiment.END_TIME ]:
                    v = dateutil.parser.parse(v)    # patch ISO-format strings to datetime objects
                params[n] = v
            self.addPendingResult(params, jobid)

    def _newload(self, j : Dict[str, Any]):
        '''Load a new-format file.

        In this format we save everything as dicts, nested or otherwise.

        :param j: the JSON object'''

        # notebook-level metadata
        self.setDescription(j['description'])
        currentTag = j['current']

        # result sets
        for tag in j['resultsets']:
            rs = self.addResultSet(tag)
            res = j['resultsets'][tag]

            # attributes
            locked = False
            for k in res.keys():
                if k == 'results':
                    # results, deal with them next
                    pass
                elif k == 'description':
                    # description handled separately
                    rs.setDescription(res[k])
                elif k == 'locked':
                    # locked flag
                    locked = res[k]
                else:
                    rs[k] = res[k]

            # results
            rcs = res['results']
            for rc in rcs:
                meta = rc[Experiment.METADATA]
                for k in meta:
                    if k in [ Experiment.START_TIME, Experiment.END_TIME ]:
                        try:
                            # patch ISO-format strings to datetime objects
                            meta[k] = dateutil.parser.parse(meta[k]) 
                        except:
                            # leave unchanged
                            pass
                self.addResult(rc, tag=tag)

            # pending results
            pendings = res['pending-results']
            for jobid in pendings:
                params = dict(pendings[jobid])
                self.addPendingResult(params, jobid, tag=tag)

            # lock the result set if flagged
            if locked:
                rs.finish()

        # select the correct result set
        self.select(currentTag)

        # lock the notebook if needed
        if j['locked']:
            self.finish()

    def _save( self, fn ):
        """Persist the notebook to the given file. Note that, while we can load
        both old and new formats, we only save in the new format.  

        :param fn: the file name"""

        # result sets
        rsrcs = dict()
        for tag in self.resultSets():
            rs = self.resultSet(tag)

            # results (as nested results dicts)
            rsres = dict()
            rsres['results'] = list(rs.results())

            # pending results
            pending = dict()
            for jobid in rs.pendingResults():
                params = rs.pendingResultParameters(jobid)
                pending[jobid] = params
            rsres['pending-results'] = pending

            # attributes
            for k in rs.keys():
                rsres[k] = rs[k]

            # description
            rsres['description'] = rs.description()

            # lock
            rsres['locked'] = rs.isLocked()

            # store the whole result set
            rsrcs[tag] = rsres

        # create the JSON object
        j = json.dumps({ 'creator': PackageContactInfo,
                         'version': '2',
                         'description': self.description(),
                         'current' : self.currentTag(),
                         'locked' : self.isLocked(),
                         'resultsets': rsrcs },
                       indent = 4,
                       cls = MetadataEncoder)

        # write to file
        with open(fn, 'w') as f:
            f.write(j)
            



# Simulation "lab" experiment management, parallel cluster version
#
# Copyright (C) 2016--2020 Simon Dobson
# 
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import Lab, LabNotebook, Experiment
import numpy                                  # type: ignore
import time
import sys
from ipyparallel import Client, DirectView    # type: ignore
from contextlib import AbstractContextManager

class ClusterLab(Lab):
    """A :class:`Lab` running on an ``pyparallel`` compute cluster.

    Experiments are submitted to engines in the cluster for
    execution in parallel, with the experiments being performed
    asynchronously to allow for disconnection and subsequent retrieval
    of results. Combined with a persistent :class:`LabNotebook`, this allows
    for fully decoupled access to an on-going computational experiment
    with piecewise retrieval of results.

    This class requires a cluster to already be set up and running, configured
    for persistent access, with access to the necessary code and libraries,
    and with appropriate security information available to the client.
    """

    # Tuning parameters
    WaitingTime : int = 30           #: Waiting time for checking for job completion. Lower values increase network traffic.
    Reconnections : int = 5          #: Number of attempts when re-connecting to a cluster.
    Retries : int = 3                #: Number of re-tries for failed jobs.
    
    def __init__( self, notebook : LabNotebook = None, url_file = None, profile = None, profile_dir = None, ipython_dir = None, context = None, debug = False, sshserver = None, sshkey = None, password = None, paramiko = None, timeout = 10, cluster_id = None, **extra_args ):
        """Create an empty lab attached to the given cluster. Most of the arguments
        are as expected by the ``pyparallel.Client`` class, and are used to create the
        underlying connection to the cluster. The connection is opened immediately,
        meaning the cluster must be up and accessible when creating a lab to use it.

        :param notebook: the notebook used to results (defaults to an empty :class:`LabNotebook`)
        :param url_file: file containing connection information for accessing cluster
        :param profile: name of the IPython profile to use
        :param profile_dir: directory containing the profile's connection information
        :param ipython_dir: directory containing profile directories
        :param context: ZMQ context
        :param debug: whether to issue debugging information (defaults to False)
        :param sshserver: username and machine for ssh connections
        :param sshkey: file containing ssh key
        :param password: ssh password
        :param paramiko: True to use paramiko for ssh (defaults to False)
        :param timeout: timeout in seconds for ssh connection (defaults to 10s)
        :param cluster_id: string added to runtime files to prevent collisions"""
        super(ClusterLab, self).__init__(notebook)
        
        # record all the connection arguments for later
        self._arguments = dict(url_file = url_file,
                               profile = profile,
                               profile_dir = profile_dir,
                               ipython_dir = ipython_dir,
                               context = context,
                               debug = debug,
                               sshserver = sshserver,
                               sshkey = sshkey,
                               password = password,
                               paramiko = paramiko,
                               timeout = timeout,
                               cluster_id = cluster_id,
                               **extra_args)
        self._client : Client = None

        # connect to the cluster
        self.open()

        # use the more sophisticated pickler
        rc = self._client[:].use_cloudpickle()


    # ---------- Protocol ----------

    def connect(self):
        '''Low-level connection to the cluster. Most code should
        use :meth:`open` to open the connection: this method
        performs a single connection attempt, raising an exception if it
        fails.'''
        if self._client is None:
            self._client = Client(**self._arguments)

            # if we get here, activate the connection
            self.activate()

    def activate(self):
        '''Make the connection active to ipyparallel/Jupyter. Usually
        only needed when there are several labs active in one program,
        where this method selects the lab used by, fo example, parallel magics.'''
        if self._client is None:
            raise Exception('No open connection to activate')
        else:
            self._client.direct_view().activate()

    def open(self):
        """Connect to the cluster. This will involve several possible re-tries."""
        try:
            # first try
            if self._client is None:
                # no open connection, try to connect
                self.connect()
            else:
                # we have an open connection, activate it
                self.activate()
        except Exception as exc:
            # if we get here, we're definitely disconnected, so try
            # to re-connect the requisite number of times
            for i in range(self.Reconnections):
                print('Connection to cluster failed, reconnecting ({i}/{n})'.format(i=i + 1, n=self.Reconnections), file=sys.stderr)
                try:
                    # try to connect
                    self.connect()

                    # if we get here, we succeeded
                    return
                except Exception as e:
                    # go round the loop again, saving the exception in
                    # case we're the last try
                    exc = e
            
            # OK, we've failed enough, stop punching the brick wall...
            raise exc

    def close( self ):
        """Close down the connection to the cluster."""
        if self._client is not None:
            self._client.close()
            self._client = None

    def recreate(self):
        '''Save the arguments needed to re-connect to the cluster we use.

        :returns: a (classname, args) pair'''
        (_, args) = super(ClusterLab, self).recreate()
        n = '{modulename}.{classname}'.format(modulename = self.__class__.__module__,
                                              classname = self.__class__.__name__)
        nargs = args.copy()
        nargs.update(self._arguments)
        return (n, nargs)


    # ---------- Remote control of the compute engines ----------

    def numberOfEngines(self) -> int:
        """Return the number of engines available to this lab.

        :returns: the number of engines"""
        return len(self.engines())

    def engines(self) -> DirectView:
        """Return a list of the available engines.

        :returns: a list of engines"""
        self.open()
        return self._client[:]

    def sync_imports(self, quiet : bool =False) -> AbstractContextManager:
        """Return a context manager to control imports onto all the engines
        in the underlying cluster. This method is used within a ``with`` statement.

        Any imports should be done with no experiments running, otherwise the
        method will block until the cluster is quiet. Generally imports will be one
        of the first things done when connecting to a cluster. (But be careful
        not to accidentally try to re-import if re-connecting to a running
        cluster.)

        :param quiet: if True, suppresses messages (defaults to False)
        :returns: a context manager"""
        self.open()
        return self._client[:].sync_imports(quiet=quiet)


    # ---------- Running experiments ----------

    def runExperiment(self, e : Experiment):
        """Run the experiment across the parameter space in parallel using
        all the engines in the cluster. This method returns immediately.

        The experiments are run asynchronously, with the points in the parameter
        space being explored randomly so that intermediate retrievals of results
        are more representative of the overall result. Put another way, for a lot
        of experiments the results available will converge towards a final
        answer, so we can plot them and see the answer emerge.

        :param e: the experiment"""

        # create the parameter space
        space = self.parameterSpace()

        # only proceed if there's work to do
        if len(space) > 0:
            nb = self.notebook()
           
            # randomise the order of the parameter space so that we evaluate across
            # the space as we go along to try to make intermediate (incomplete) result
            # sets more representative of the overall result set
            ps = space.copy()
            numpy.random.shuffle(ps)

            try:
                # connect to the cluster
                self.open()

                # configure a load balanced view of the cluster
                view = self._client.load_balanced_view()
                view.set_flags(retries=self.Retries)

                # submit an experiment at each point in the parameter space to the cluster
                jobs = []
                for p in ps:
                    rc = view.apply_async(lambda ep: ep[0].set(ep[1]).run(), (e, p))
                    ids = rc.msg_ids
                    #print('Started {ids}'.format(ids=ids))
                    jobs.extend(ids)

                    # there seems to be a race condition in submitting jobs,
                    # whereby jobs get dropped if they're submitted too quickly
                    time.sleep(0.01)
                
                # record the mesage ids of all the jobs as submitted but not yet completed
                psjs = zip(ps, jobs)
                for (p, j) in psjs:
                    nb.addPendingResult(p, j)
            finally:
                # commit our pending results in the notebook
                nb.commit()
                self.close()

    def updateResults(self, purge : bool =False) -> int:
        """Update our results within any pending results that have completed since we
        last retrieved results from the cluster. Optionally purges any jobs that
        have crashed, which can be due to engine failure within the
        cluster. This prevents individual crashes blocking the retrieval of other jobs.

        :param purge: (optional) cancel any jobs that have crashed (defaults to False)
        :returns: the number of pending results completed at this call"""
        nb = self.notebook()

        # look for pending results if we're waiting for any
        n = 0
        if not nb.ready():
            # we have results to get
            try:
                crashed = []
                self.open()
                for j in set(nb.allPendingResults()):
                    try:
                        # query the status of a job
                        #print('Test status of {j}'.format(j=j))
                        status = self._client.result_status(j, status_only=False)
                    
                        # add all completed jobs to the notebook
                        if j in status['completed']:
                            r = status[j]
                            #print('{j} completed'.format(j=j))

                            # resolve the result in the appropriate result set
                            nb.resolvePendingResult(r, j)

                            # record that we retrieved the results for the given job
                            n = n + 1
                    
                            # purge the completed job from the cluster
                            self._client.purge_hub_results(j)        
                    except Exception as e:
                        # report the exception and carry on, recording the job as crashed
                        print(e, file=sys.stderr)
                        crashed.append(j)

                # purge any crashed jobs if requested
                if purge and len(crashed) > 0:
                    for j in crashed:
                        nb.cancelPendingResult(j)
                        self._client.purge_hub_results(j)    
            finally:
                # commit changes to the notebook and close the connection
                nb.commit()
                self.close()

        return n


    # ---------- Accessing results ----------
    
    def wait(self, timeout : int =-1) -> bool:
        """Wait for all pending results in all result sets to be finished. If timeout is set,
        return after this many seconds regardless.

        :param timeout: timeout period in seconds (defaults to forever)
        :returns: True if all the results completed"""

        # we can't use pyparallel.Client.wait() for this, because that
        # method only works for cases where the Client object is the one that
        # submitted the jobs to the cluster hub -- and therefore has the
        # necessary data structures to perform synchronisation. This isn't the
        # case for us, as one of the main goals of epyc is to support disconnected
        # operation, which implies a different Client object retrieving results
        # than the one that submitted the jobs in the first place. This is
        # unfortunate, but understandable given the typical use cases for
        # Client objects in pyparallel.
        #
        # Instead. we have to code around a little busily. The ClusterLab.WaitingTime
        # global sets the latency for waiting, and we repeatedly wait for this amount
        # of time before updating the results. The latency value essentially controls
        # how busy this process is: given that most simulations are expected to
        # be long, a latency in the tens of seconds feels about right as a default
        nb = self.notebook()
        if nb.numberOfAllPendingResults() > 0:
            # we've got pending results, wait for them
            timeWaited = 0
            while (timeout < 0) or (timeWaited < timeout):
                self.updateResults()
                if nb.numberOfAllPendingResults() == 0:
                    # no pending results left, we're complete
                    return True
                else:
                    # not done yet, calculate the waiting period
                    if timeout == -1:
                        # wait for the default waiting period
                        dt = self.WaitingTime
                    else:
                        # wait for the default waiting period or until the end of the timeout.
                        # whichever comes first
                        if (timeout - timeWaited) < self.WaitingTime:
                            dt = timeout - timeWaited
                        else:
                            dt = self.WaitingTime
                            
                    # sleep for a while
                    time.sleep(dt)
                    timeWaited = timeWaited + dt

            # if we get here, the timeout expired, so do a final check
            # and then exit
            return (nb.numberOfAllPendingResults() == 0)

        else:
            # no results, so we got them all
            return True

       



# Initialisation for epyc package
#
# Copyright (C) 2016--2020 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

"""`epyc` is a Python module for controlling a long-running series of
computational experiments, as is often found when writing simulations
of complex networks and other such domains. There is often a need to
perform a computation across a multi-dimensional parameter space,
varying the parameters, performing and aggregating multiple
repetitions, and wrangling results for analysis and
presentation. Often the experiments being performed are on such a
scale as to require the use of a computing cluster to perform multiple
experiments simultaneously.

Managing all these tasks is complicated, so `epyc` tries to automate
the process. It provides a way to define a :term:`lab` that performs
an :term:`experiment` (or, more likely, a sequence of experiments)
whose parameters and results are recorded as a :term:`result set` in a lab :term:`notebook`
for later retrieval. Laboratories can be sequential (for a single
machine) or parallel (to use a multicore or cluster of machines); lab
notebooks can be persistent to allow experiments to be fired-off and
their results retrieved later -- handy if you use a laptop. Notebooks
store all the data and metadata immutably in a portable format
(JSON or HDF5) to improve the reproducibility of computational experiments.

`epyc` also includes a small number of :term:`experiment combinators` that
separate the logic of a single experiment from the logic of performing
multiple repetitions and other structuring tasks. This means that any
experiment can be repeated and statistically summarised, for example.

"""

# String written into every persistent notebook file
PackageContactInfo = 'Created by epyc, computational experiment management for Python <https://pypi.org/project/epyc/>'

from .experiment import Experiment, ResultsDict
from .experimentcombinator import ExperimentCombinator
from .repeatedexperiment import RepeatedExperiment
from .summaryexperiment import SummaryExperiment

from .resultset import ResultSet, ResultSetLockedException, CancelledException, PendingResultException
from .labnotebook import LabNotebook, ResultsStructureException, NotebookVersionException, LabNotebookLockedException
from .jsonlabnotebook import JSONLabNotebook
from .hdf5labnotebook import HDF5LabNotebook

from .lab import Lab
from .parallellab import ParallelLab
from .clusterlab import ClusterLab

# Late and/or complex initialisation
Experiment._init_statics()
ResultSet._init_statics()





# Base class for experiment combinators that co=ordinate other experiments
#
# Copyright (C) 2016--2020 Simon Dobson
# 
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import Experiment
from typing import Dict, Any


class ExperimentCombinator(Experiment):
    """An experiment that wraps-up another, underlying experiment. This is an abstract
    class that just provides the common wrapping logic.

    Experiment combinators aren't expected to have parameters of their own: they
    simply use the parameters of their underlying experiment. They may however
    give rise to metadata of their own, and modify the results returned by running
    their underlying experiment."""

    def __init__(self, ex : Experiment):
        """Create a combinator based on the given experiment.

        ex: the underlying experiment"""
        super(ExperimentCombinator, self).__init__()
        self._experiment = ex

    def experiment(self) -> Experiment:
        """Return the underlying experiment.

        :returns: the underlying experiment"""
        return self._experiment

    def set(self, params : Dict[str, Any]) -> Experiment:
        """Set the parameters for the experiment, returning the
        now-configured experiment.

        :param params: the parameters
        :returns: the experiment combinator itself"""
        self.experiment().set(params)
        return self

    def parameters(self) -> Dict[str, Any]:
        """Return the current experimental parameters, taken from the
        underlying experiment.

        :returns: the parameters,"""
        return self.experiment().parameters()

    



# Simulation "lab notebook" for collecting results, in-memory version
#
# Copyright (C) 2016--2020 Simon Dobson
#
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import Experiment, ResultSet, ResultsDict
from pandas import DataFrame                               # type: ignore
from contextlib import contextmanager
import sys
if sys.version_info >= (3, 8):
    from typing import List, Set, Dict, Any, Optional, Union, Final, cast
else:
    # backwards compatibility with Python 35, Python36, and Python37 
    from typing import List, Set, Dict, Any, Optional, Union, cast
    from typing_extensions import Final


class ResultsStructureException(Exception):
    '''An exception raised when there is a problem with the structure of a
    results dict. 

    :param rc: the results dict structure that causes the problem''' 

    def __init__(self, rc : Union[ResultsDict, List[ResultsDict]]):
        super(ResultsStructureException, self).__init__("Can't handle results dict {rc}".format(rc=rc))
        self._rc = rc

    def resultsdict(self) -> Union[ResultsDict, List[ResultsDict]]:
        '''Return the results dict that caused the problem.

        :returns: the results dict or list of them'''
        return self._rc


class NotebookVersionException(Exception):
    '''An exception raised when a notebook encounters an unexpected version
    of a persistent file format.

    :param expected: the expected version
    :param actual: the actual version'''

    def __init__(self, expected : str, actual : str):
        super(NotebookVersionException, self).__init__('Expected notebook version {v1}, got {v2}'.format(v1=expected, v2=actual))
        self._expected = expected
        self._actual = actual
    
    def expectedVersion(self) -> str:
        '''Return the expected version of the file.

        :returns: the expected version'''
        return self._expected
    
    def actualVersion(self) -> str:
        '''Return the actual version of the file.

        :returns: the actual version'''
        return self._expected

class LabNotebookLockedException(Exception):
    '''An exception raised if an attempt is made to write to a notebook
    that's been locked by a call to :meth:`LabNotebook.finish`. This includes
    attemoting to add result sets.'''

    def __init__(self):
        super(LabNotebookLockedException, self).__init__('Lab notebook locked')



class LabNotebook(object):
    '''A "laboratory notebook" collecting together the results obtained from
    different sets of experiments. A notebook is composed of :class:`ResultSet` objects,
    which are homogeneous collections of results of experiments performed at different values
    for the same set of parameters. Each result set is tagged for access,
    with the notebook using one result set as "current" at any time.

    The notebook collects together pending results from all result sets so that they
    can be accessed uniformly. This is used by labs to resolve pending results if
    there are multiple sets of experiments running simultaneously.

    :param name: (optional) the notebook name (may be meaningful for sub-classes)
    :param description: (optional) a free text description'''
 
    # Defaults
    DEFAULT_RESULTSET : Final[str] = 'epyc.resultset.default'  #: Tag for the default result set.

    def __init__(self, name : str ='', description : str =None):
        if description is None:
            description = 'A notebook' 
        self._name : Optional[str] = name                    # name
        self._description : str = description                # description
        self._resultSets: Dict[str, ResultSet] = dict()      # tag to result set
        self._resultSetTags : Dict[ResultSet, str] = dict()  # result set to tag
        self._pending : Dict[str, ResultSet] = dict()        # pending results job ids to result sets
        self._locked : bool = False                          # locked flag

        # add a result set with the default tag, and make it current
        defrc = self.addResultSet(self.DEFAULT_RESULTSET)
        self._current : ResultSet = defrc
        #defrc.dirty(False)                  # default shouldn't trigger a write


    # ---------- Access ----------

    def name(self) -> str:
        """Return the name of the notebook. If the notebook is persistent,
        this likely relates to its storage in some way (for example a
        file name).

        :returns: the notebook name or None"""
        return self._name

    def description(self) -> str:
        """Return the free text description of the notebook.

        :returns: the notebook description"""
        return self._description

    def setDescription(self, d : str):
        '''Set the free text description of the notebook.

        :param d: the description'''
        self.assertUnlocked()
        self._description = d


    # ---------- Persistence ----------

    def isPersistent(self) -> bool:
        """By default notebooks are not persistent.

        :returns: False"""
        return False

    def commit(self):
        """Commit to persistent storage. By default does nothing. This should
        be called periodically to save intermediate results: it may happen
        automatically in some sub-classes, depending on their implementation."""
        pass

    def _commit(self):
        '''Private API for forcing commits, called when finishing a notebook.
        Defaults to the same behaviour as  :meth:`commit`.'''
        self.commit()


    # ---------- Managing results sets ----------

    def addResultSet(self, tag : str, description : str =None) -> ResultSet:
        '''Start a new experiment. This creates a new result set to hold
        the results, which will receive any results and notes.

        :param tag: unique tag for this result set
        :param: (optional) free text description of the result set
        :returns: the result set'''
        self.assertUnlocked()
        rs = ResultSet(description)
        self._resultSets[tag] = rs
        self._resultSetTags[rs] = tag
        self._current = rs
        return self._current

    def resultSets(self) -> List[str]:
        '''Return the tags for all the result sets in this notebook.

        :returns: a list of keys'''
        return list(self._resultSets.keys())

    def resultSet(self, tag : str) -> ResultSet:
        '''Return the tagged result set.

        :param tag: the tag
        :returns: the result set'''
        return self._resultSets[tag]

    def resultSetTag(self, rs : ResultSet) -> str:
        '''Return the tag associated with the given result set.

        :param rs: the result set
        :returns: the tag'''
        return self._resultSetTags[rs] 

    def select(self, tag: str) -> ResultSet:
        '''Select the given result set as current. Sub-classes may use this
        to manage memory, for example by swapping-out non-current result sets.

        :param tag: the tag
        :returns: the result set'''
        self._current = self._resultSets[tag] 
        return self._current

    def current(self) -> ResultSet:
        '''Return the current result set.

        :returns: the result set'''
        return self._current

    def currentTag(self) -> str:
        '''Return the tag of the current result set.

        :returns: the tag'''
        return self._resultSetTags[self._current]

    def numberOfResultSets(self) -> int:
        '''Return the number of result sets in this notebook.

        :returns: the number of result sets'''
        return len(self._resultSets)

    def __len__(self) -> int:
        '''Return the number of result sets in this notebook.
        Same as :meth:`numberOfResultSets`.

        :returns: the number of result sets'''
        return self.numberOfResultSets()


    # ---------- Finishing ----------

    def finish(self, commit=True):
        '''Mark the entire notebook as finished, closing and locking all result sets
        against further changes. Finishing a persistent notebook commits it.
        
        By default the finished notebook is committed as such. In certain cases
        it may be desirable to finish the notebook but not commit it, *i.e.*, to
        stop updates in memory without changing the backing file. Setting ``commit=False``
        will accomplish this.
        
        :param commit: (optional) commit the notebook (defaults to True)'''
        if not self.isLocked():
            for tag in self.resultSets():
                rs = self.resultSet(tag)
                rs.finish()
            self._locked = True   # set here to make sure the status is committed
            if commit:
                self._commit()

    def isLocked(self) -> bool:
        '''Returns true if the notebook is locked.

        :returns: True if the notebook is locked'''
        return self._locked

    def assertUnlocked(self):
        '''Tests whether the notebook is locked, and raises a :class:`LabNotebookLockedException`
        if so.'''
        if self.isLocked():
            raise LabNotebookLockedException()


    # ---------- Managing pending results in the current result set ----------

    def addPendingResult(self, params : Dict[str, Any], jobid : str, tag : str =None):
        '''Add a pending result for the given point in the parameter space
        under the given job identifier to the current result set. The identifier
        will generally be meaningful to the lab that submitted the request, and
        must be unique.

        :param params: the experimental parameters
        :param jobid: the job id
        :param tag: (optional) the tag of the result set receiving the pending result (defaults to the current result set)'''
        self.assertUnlocked()
        if tag is None:
            rs = self._current
        else:
            rs = self._resultSets[tag]

        # add the pending job to the result set
        rs.addSinglePendingResult(params, jobid)

        # record the result set that holds the given job
        self._pending[jobid] = rs


    # ---------- Resolving and cancelling results in any result set ----------

    def resolvePendingResult(self, rc : ResultsDict, jobid : str):
        '''Resolve the pending result with the given job id with the given
        results dict. The experimental parameters of the result are sanity-checked
        against what the result set expected for that job.

        The result may not be pending within the current result set, but can
        be within any result set in the notebook. This will not affect
        the result set that is selected as current.

        :param rc: the results dict
        :param jobid: the job id'''
        self.assertUnlocked()

        # add the result to the correct result set
        tag = self._resultSetTags[self._pending[jobid]]
        rs = self._pending[jobid]
        self.addResult(rc, tag=tag)

        # resolve the pending result in that result set
        rs.resolveSinglePendingResult(jobid)

        # mark the job as resolved with the notebook
        del self._pending[jobid] 

    def cancelPendingResult(self, jobid : str):
        '''Cancel the given pending result.

        The result may not be pending within the current result set, but can
        be within any result set in the notebook. This will not affect
        the result set that is selected as current.

        :param jobid: the job id'''
        self.assertUnlocked()

        # find the result set for the job
        rs = self._pending[jobid]

        # resolve this result
        rs.cancelSinglePendingResult(jobid)

        # mark the job as resolved with the notebook
        del self._pending[jobid] 
            
    def ready(self, tag : str =None) -> bool:
        '''Test whether the result set has pending results. 

        :params tag: (optional) the result set tag (defaults to the current set)
        :returns: True if all pending results have been resolved (or cancelled)'''
        if tag is None:
            rs = self._current
        else:
            rs = self._resultSets[tag]
        return rs.ready()

    def readyFraction(self, tag : str =None) -> float:
        """Test what fraction of results are available in the tagged result set.
        
        :params tag: (optional) the result set tag (defaults to the current set)
        :returns: the fraction of available results"""
        if tag is None:
            rsc = self._current
        else:
            rsc = self._resultSets[tag]
        rsc = self.current()
        nr = rsc.numberOfResults()
        np = rsc.numberOfPendingResults()
        tr = nr + np
        if tr == 0:
            return 1.0
        else:
            return nr / tr

    def pendingResults(self, tag : str =None) -> List[str]:
        '''Return the identifiers of the results pending in the tagged dataset.
        
        :params tag: (optional) the result set tag (defaults to the current set)
        :returns: a set of job identifiers'''
        if tag is None:
            rs = self._current
        else:
            rs = self._resultSets[tag]
        return rs.pendingResults()

    def numberOfPendingResults(self, tag : str =None) -> int:
        '''Return the number of results pending in the tagged dataset.
        
        :params tag: (optional) the result set tag (defaults to the current set)
        :returns: the number of results'''
        if tag is None:
            rs = self._current
        else:
            rs = self._resultSets[tag]
        return rs.numberOfPendingResults()

    def pendingResultParameters(self, jobid : str) -> Dict[str, Any]:
        '''Return a dict of paramneters corresponding to the given pending result.

        :param jobid: the job id
        :returns: a dict of parameter values'''
        rs = self._pending[jobid]
        return rs.pendingResultParameters(jobid)


    # ---------- Managing pending results in all result sets ----------

    def allPendingResults(self) -> Set[str]:
        '''Return the identifiers for all pending results in all result sets.

        :returns: a set of job identifiers'''
        return set(self._pending.keys())

    def numberOfAllPendingResults(self) -> int:
        '''Return the number of results pending in all result sets.

        :returns: the total number of pending results'''
        return len(self._pending) 


    # --------- Managing results ----------

    def _addResult(self, result : ResultsDict, tag : str =None):
        '''Private method to add a single result.

        :param tag: (optional) the result set to add to (defaults to the current result set)
        :param result: the results dict'''
        if tag is None:
            rs = self._current
        else:
            rs = self._resultSets[tag]
        rs.addSingleResult(result)

    def addResult(self, results : Union[ResultsDict, List[ResultsDict]], tag : str =None):
        """Add one or more results dicts to the current result set. Each should
        be a :term:`results dict` as returned from
        an instance of :class:`Experiment`, that contains metadata,
        parameters, and result.

        The results may include one or more nested results dicts, for example as
        returned by :class:`RepeatedExperiment`, whose results are a list of results
        at the same point in the parameter space. In this case the embedded
        results will themselves be unpacked and added.

        One may also add a list of results dicts, in which case they will
        be added individually.

        Any structure of results dicts that can't be handled will raise a 
        :class:`ResultsStructureException`.

        :param result: a results dict or collection of them
        :param tag: (Optional) result set to add tp (defalts to the current result set)
        """

        # deal with the different ways of presenting results to be added
        if isinstance(results, list):
            # a list, recursively add all elements
            for res in results:
                self._addResult(res, tag)
        elif isinstance(results, dict):
            # a results dict, check for nesting
            if isinstance(results[Experiment.RESULTS], list):
                # a result with embedded results, unwrap and and add them
                for res in cast(List[ResultsDict], results[Experiment.RESULTS]):
                    self._addResult(res, tag)
            else:
                # a single results dict with a single set of experimental results
                self._addResult(results, tag)

        else:
            raise ResultsStructureException(results)
    
    def dataframe(self, tag : str =None, only_successful : bool =True) -> DataFrame:
        """Return results as a ``pandas.DataFrame``. If no tag is provided,
        use the current result set. 

        If the only_successful flag is set (the default), then the DataFrame will
        only include results that completed without an exception; if it is set to
        False, the DataFrame will include all results and also the exception details.

        If you are only interested in results corresponding to some sets of parameters
        you can pre-filter the dataframe using :meth:`dataframeFor`. 

        :params tag: (optional) the tag of the result set (defaults to the currently select result set)
        :param only_successful: include only successful experiments (defaults to True)
        :returns: the parameters, results, and metadata in a DataFrame"""
        if tag is None:
            rs = self._current
        else:
            rs = self._resultSets[tag]
        df = rs.dataframe(only_successful)
        return df

    def dataframeFor(self, params : Dict[str, Any], tag : str =None, only_successful : bool =True) -> DataFrame:
        """Return results for the goven parameter values as a ``pandas.DataFrame``.
        If no tag is provided, the current result set is queried. 
        If the only_successful flag is set (the default), then the DataFrame will
        only include results that completed without an exception; if it is set to
        False, the DataFrame will include all results and also the exception details.

        :param params: the experimental parameters
        :params tag: (optional) the tag of the result set (defaults to the currently select result set)
        :param only_successful: include only successful experiments (defaults to True)
        :returns: the parameters, results, and metadata in a DataFrame"""
        if tag is None:
            rs = self._current
        else:
            rs = self._resultSets[tag]
        df = rs.dataframeFor(params, only_successful)
        return df

    def results(self, tag : str =None) -> List[ResultsDict]:
        """Return results as a list of results dicts. If no tag is provided,
        use the current result set. This is a lot slower and more memory-hungry
        than using :meth:`dataframe` (which is therefore to be preferred),
        but may be useful for small sets of results that need a more Pythonic
        interface than that provided by DataFrames. You can pre-filter the
        results dicts to those matching only some parameters combinations
        using :meth:`resultsFor`.

        :params tag: (optional) the tag of the result set (defaults to the currently select result set)
        :returns: the results dicts"""
        if tag is None:
            rs = self._current
        else:
            rs = self._resultSets[tag]
        return rs.results()

    def resultsFor(self, params, tag : str =None) -> List[ResultsDict]:
        """Return results for the given parameter values a list of results dicts. If no tag is provided,
        use the current result set. This is a lot slower and more memory-hungry
        than using :meth:`dataframeFor` (which is therefore to be preferred),
        but may be useful for small sets of results that need a more Pythonic
        interface than that provided by DataFrames.

        :param params: the experimental parameters
        :returns: results dicts"""
        if tag is None:
            rs = self._current
        else:
            rs = self._resultSets[tag]
        return rs.resultsFor(params)

    def numberOfResults(self, tag : str =None) -> int:
        '''Return the number of results in the tagged dataset.
        
        :params tag: (optional) the result set tag (defaults to the current set)
        :returns: the number of results'''
        if tag is None:
            rs = self._current
        else:
            rs = self._resultSets[tag]
        return rs.numberOfResults()


    # ---------- Context manager protocol ----------

    @contextmanager
    def open(self):
        '''Open and close the notebook using a ``with`` block. For persistent
        notebooks this will cause the notebook to be committed to persistent
        storage in a robust manner.'''
        try:
            # nothing to do, as the operations all open as required
            yield self
        finally:
            # commit any changes and close the file
            self.commit()





# Tests of cluster-driven lab class
#
# Copyright (C) 2016--2020 Simon Dobson
# 
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import *

import unittest
import numpy
import time
import os
import subprocess
from tempfile import NamedTemporaryFile

# set limit low for testing purposes
ClusterLab.WaitingTime = 10


class SampleExperiment(Experiment):
    '''A very simple experiment that adds up its parameters.'''

    def do( self, param ):
        total = 0
        for k in param:
            total = total + param[k]
        return dict(total = total)

class SampleExperiment2(Experiment):
    '''Add up after waiting.'''

    def do( self, param ):
        time.sleep(2)
        total = 0
        for k in param:
            total = total + param[k]
        return dict(total = total)

    
# use the existence of an ipcluster.pid file in the epyctest IPython
# profile directory as a proxy for there being a cluster running
# that we can use for our tests
# (cluster is created by running `make cluster` in the project Makefile)
profile = 'epyctest'
profile_dir = subprocess.check_output('ipython locate profile {p}'.format(p=profile), shell=True).rstrip().decode()
pid_file = os.path.join(profile_dir, 'pid/ipcluster.pid')
@unittest.skipUnless(os.path.isfile(pid_file),
                     "No {p} cluster running (no pid file  {fn})".format(p=profile, fn=pid_file))
class ClusterLabTests(unittest.TestCase):

    def setUp( self ):
        '''Create a lab in which to perform tests and a tempfile for HDF5 tests.'''
        self._lab = ClusterLab(profile=profile)
        #self._lab = ClusterLab(url_file='ipcontroller-client.json')
        tf = NamedTemporaryFile()
        tf.close()
        self._fn = tf.name
        #self._fn = 'test.h5'

    def tearDown( self ):
        '''Close the conection to the cluster and delete the test HDF5 file.'''
        self._lab.close()
        self._lab = None
        try:
            os.remove(self._fn)
            #pass
        except OSError:
            pass

    def _results(self):
        '''Retrieve results as a list of dicts, for easier comparisons.'''
        df = self._lab.dataframe()
        res = []
        for i in df.index:
            row = dict()
            for k in df.columns:
                row[k] = df.loc[i][k]
            res.append(row)
        return res

    def testEmpty( self ):
        '''Test that things work for an empty lab'''
        self.assertEqual(self._lab.notebook().readyFraction(), 1.0)
        self.assertEqual(len(self._lab.dataframe()), 0)
        
    def testRunExprimentSync( self ):
        '''Test running an experiment and grabbing all the results by sleeping for a while.'''
        n = 20

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment())
        time.sleep(n * 2.5 / self._lab.numberOfEngines())
        self.assertTrue(self._lab.ready())
        res = self._results()
        
        # check that the whole parameter space has a result
        self.assertEqual(len(res), n)
        for p in res:
            self.assertIn(p['a'], r)

        # check that each result corresponds to its parameter
        for p in res:
            self.assertEqual(p['a'], p['total'])
            
    def testWait( self ):
        '''Test waiting for all jobs to complete.'''
        n = 20

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment())
        self.assertTrue(self._lab.wait())
        self.assertTrue(self._lab.ready())
            
    def testWaitShortTimeout( self ):
        '''Test short-timeout (and short-latency) waiting.'''
        n = 20
        self._lab.WaitingTime = 5
        
        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment2())
        self.assertFalse(self._lab.wait(timeout=5))
        self.assertFalse(self._lab.ready())
        self.assertTrue(self._lab.wait())
        self.assertTrue(self._lab.ready())

    def testRunExprimentAsync( self ):
        '''Test running an experiment and check the results come in piecemeal.'''
        n = 20

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment())

        # watch results coming in
        f = 0.0
        while f < 1:
            f1 = self._lab.readyFraction()
            #print self._lab._availableResults(), f1
            self.assertTrue(f1 >= f)
            f = f1
        self.assertTrue(self._lab.ready())
        self.assertEqual(self._lab.notebook().numberOfPendingResults(), 0)
        res = self._results()
        
        # check that the whole parameter space has a result
        self.assertEqual(len(res), n)
        for p in res:
            self.assertIn(p['a'], r)

        # check that each result corresponds to its parameter
        for p in res:
            self.assertEqual(p['a'], p['total'])

    def testReturnWithNoJobs( self ):
        '''Test wait() returns True when there are no jobs pending.'''
        n = 20
          
        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment())
        self.assertTrue(self._lab.wait())
        
        # calling wait() again should also be true (with no delay, which we don't check for)
        self.assertTrue(self._lab.wait())

    def testCancelSomeJobs( self ):
        '''Test we can cancel some jobs while keeping the rest.'''
        n = 20
          
        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment2())

        # cancel a job, check we retained them all
        params = dict(a = int(n / 2))
        jobids = self._lab.notebook().current().pendingResultsFor(params)
        for j in jobids:
            self._lab.notebook().cancelPendingResult(j)
        self._lab.wait()
        self.assertEqual(self._lab.notebook().current().numberOfResults(), n)
        self.assertEqual(self._lab.notebook().current().numberOfPendingResults(), 0)

        # check cancelled job had a marker exception
        for rc in self._lab.notebook().results():
            if rc[Experiment.PARAMETERS]['a'] == int(n / 2):
                self.assertTrue(isinstance(rc[Experiment.METADATA][Experiment.EXCEPTION], CancelledException))


    def testCancelAllJobs( self ):
        '''Test we can cancel all jobs.'''
        n = 20
          
        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment2())

        # cancel all jobs
        for j in self._lab.notebook().allPendingResults():
            self._lab.notebook().cancelPendingResult(j)
        self._lab.wait()
        self.assertEqual(self._lab.notebook().current().numberOfResults(), 20)
        self.assertEqual(self._lab.notebook().current().numberOfPendingResults(), 0)

        # make sure all jobs have a marker
        for rc in self._lab.notebook().results():
            self.assertTrue(isinstance(rc[Experiment.METADATA][Experiment.EXCEPTION], CancelledException))


    def testAddExperiments( self ):
        '''Test we can add experiments while some are running, without locking up.'''
        n = 20

        # run the first experiment
        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment2())

        # while this is waiting, run another
        r = numpy.arange(n, 2 * n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment2())

        self._lab.wait()
        self.assertEqual(len(self._lab.notebook().dataframe()), 2 * n)
        self.assertEqual(self._lab.notebook().current().numberOfPendingResults(), 0)

    def testClusterAndHDF5(self):
        '''Test that a cluster works with HDF5 properly.'''
        self._lab = ClusterLab(profile=profile,
                               notebook=HDF5LabNotebook(self._fn, create=True))
        self._lab.notebook().addResultSet('mydata')

        # run an experiment, with all the trimmings
        n = 20
        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment2())
        self._lab.wait()
        self._lab.notebook().commit()

        with self._lab.notebook().open() as nb:
            self.assertEqual(len(nb), 2)         # number of result sets
            self.assertEqual(len(nb.current()), n)
            self.assertEqual(nb.numberOfResults(), n)
            self.assertEqual(nb.numberOfPendingResults(), 0)


# TODO: Test we can run pending jobs from different result sets

if __name__ == '__main__':
    unittest.main()

       



# Tests of summarising experiments combinator
#
# Copyright (C) 2016--2020 Simon Dobson
# 
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.


from epyc import *

import unittest
import numpy
import os
from tempfile import NamedTemporaryFile


class SampleExperiment1(Experiment):
    
    def do( self, params ):
        return dict(result = params['x'],
                    dummy = 1)

class SampleExperiment2(Experiment):

    def __init__( self ):
        super(SampleExperiment2, self).__init__()
        self._rng = numpy.random.default_rng()
        self._allvalues = []
    
    def do( self, params ):
        v = self._rng.random()
        self._allvalues.append(v)
        return dict(result = v)

    def values( self ):
        return self._allvalues

class SampleExperiment3(SampleExperiment2):

    def __init__( self ):
        super(SampleExperiment3, self).__init__()
        self._ran = 0
    
    def do( self, params ):
        v = self._rng.random()
        if v < 0.5:
            # experiment succeeds
            self._allvalues.append(v)
            self._ran = self._ran + 1
            return dict(result = v)
        else:
            # experiment fails
            raise Exception("Failing")

    def ran( self ):
        return self._ran

class SampleExperiment4(Experiment):
    '''A more exercising experiment.'''

    def do( self, param ):
        total = 0
        for k in param:
            total = total + param[k]
        return dict(total = total,
                    nestedArray = [ 1, 2, 3 ],
                    nestedDict = dict(a = 1, b = 'two'))

class SampleExperiment5(Experiment):

    def __init__( self ):
        super(SampleExperiment5, self).__init__()
        self._rng = numpy.random.default_rng()
        self._values = []

    def do( self, params ):
        v = self._rng.random()
        self._values.append(v)
        return dict(result = v)
    
    
class SummaryExperimentTests(unittest.TestCase):

    def setUp( self ):
        self._lab = Lab()
        
    def testRepetitionsOnePoint( self ):
        '''Test we can repeat an experiment at a single point'''
        N = 10
        
        self._lab['x'] = [ 5 ]
        
        e = SampleExperiment1()
        es = SummaryExperiment(RepeatedExperiment(e, N))

        self._lab.runExperiment(es)
        self.assertTrue(es.success())
        res = (self._lab.results())[0]
        
        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])
        self.assertEqual(res[Experiment.METADATA][SummaryExperiment.UNDERLYING_RESULTS], N)
        
    def testRepetitionsMultiplePoint( self ):
        '''Test we can repeat an experiment across a parameter space'''
        N = 10
        
        self._lab['x'] = [ 5, 10, 15 ]
        
        e = SampleExperiment1()
        es = SummaryExperiment(RepeatedExperiment(e, N))

        self._lab.runExperiment(es)
        self.assertTrue(es.success())
        res = self._lab.results()
        self.assertEqual(len(res), len(self._lab['x']))

        for i in range(len(self._lab['x'])):
            self.assertTrue(res[i][Experiment.METADATA][Experiment.STATUS])
            self.assertIn(res[i][Experiment.PARAMETERS]['x'], self._lab['x'])
            self.assertEqual(res[i][Experiment.METADATA][SummaryExperiment.UNDERLYING_RESULTS], N)
            self.assertEqual(res[i][Experiment.RESULTS]['result_mean'],
                             res[i][Experiment.PARAMETERS]['x'])
            self.assertEqual(res[i][Experiment.RESULTS]['result_variance'], 0)

    def testSummary( self ):
        '''Test that summary statistics are created properly'''
        N = 10
        
        self._lab['x'] = [ 5 ]
        
        e = SampleExperiment2()
        es = SummaryExperiment(RepeatedExperiment(e, N))

        self._lab.runExperiment(es)
        self.assertTrue(es.success())
        res = (self._lab.results())[0]

        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])
        self.assertEqual(res[Experiment.RESULTS]['result_mean'],
                         numpy.mean(e.values()))
        self.assertEqual(res[Experiment.RESULTS]['result_variance'],
                         numpy.var(e.values()))
       
    def testSelect( self ):
        '''Test we can select summary results'''
        N = 10
        
        self._lab['x'] = [ 5 ]
        
        e = SampleExperiment1()
        es = SummaryExperiment(RepeatedExperiment(e, N),
                               summarised_results = [ 'dummy' ])

        self._lab.runExperiment(es)
        self.assertTrue(es.success())
        res = (self._lab.results())[0]

        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])
        self.assertCountEqual(res[Experiment.RESULTS].keys(), [ 'dummy_mean',
                                                                'dummy_median',
                                                                'dummy_variance',
                                                                'dummy_min',
                                                                'dummy_max' ])
        self.assertEqual(res[Experiment.RESULTS]['dummy_mean'], 1)
        self.assertEqual(res[Experiment.RESULTS]['dummy_median'], 1)
        self.assertEqual(res[Experiment.RESULTS]['dummy_variance'], 0)
        self.assertEqual(res[Experiment.RESULTS]['dummy_min'], 1)
        self.assertEqual(res[Experiment.RESULTS]['dummy_max'], 1)

    def testIgnoring( self ):
        '''Test that we ignore failed experiments'''
        N = 10
        
        self._lab['x'] = [ 5 ]
        
        e = SampleExperiment3()
        es = SummaryExperiment(RepeatedExperiment(e, N))

        self._lab.runExperiment(es)
        self.assertTrue(es.success())
        res = (self._lab.results())[0]
        
        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])
        self.assertEqual(res[Experiment.METADATA][SummaryExperiment.UNDERLYING_RESULTS], N)
        self.assertEqual(res[Experiment.METADATA][SummaryExperiment.UNDERLYING_SUCCESSFUL_RESULTS], e.ran())
        self.assertEqual(res[Experiment.RESULTS]['result_mean'],
                         numpy.mean(e.values()))
        self.assertEqual(res[Experiment.RESULTS]['result_variance'],
                         numpy.var(e.values()))
         
    def testSingle( self ):
        '''Test against a non-list-returning underlying experiment.'''

        self._lab['x'] = [ 5 ]
        
        e = SampleExperiment1()
        es = SummaryExperiment(e)

        self._lab.runExperiment(es)
        self.assertTrue(es.success())
        self.assertEqual(len(self._lab.results()), 1)
        res = (self._lab.results())[0]
        
        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])
        self.assertEqual(res[Experiment.METADATA][SummaryExperiment.UNDERLYING_RESULTS], 1)
        self.assertEqual(res[Experiment.METADATA][SummaryExperiment.UNDERLYING_SUCCESSFUL_RESULTS], 1)

    def testJSON( self ):
        '''Test that we can persist repeated results in JSON format.'''

        # reset the lab we're using to use a JSON notebook
        tf = NamedTemporaryFile()
        tf.close()
        self._lab = Lab(notebook = JSONLabNotebook(tf.name, create = True))
        
        repetitions = 5
        self._lab['a'] = [ 1, 2, 3 ]
        try:
            re = RepeatedExperiment(SampleExperiment4(),
                                    repetitions)
            self._lab.runExperiment(SummaryExperiment(re,
                                                      summarised_results = [ 'total', 'nestedArray' ]))

            self.assertTrue(re.success())
            res = self._lab.results()

            # getting here is enough to exercise the persistence regime
        finally:
            try:
                os.remove(tf.name)
            except OSError:
                pass

    def testSummaryExceptions( self ):
        '''Test we handle illegal summary keys.'''
        N = 10
        
        self._lab['x'] = 'hello'
        
        e = SampleExperiment1()
        es = SummaryExperiment(RepeatedExperiment(e, N))

        self._lab.runExperiment(es)
        self.assertEqual(len(self._lab.results()), 1)
        rc = (self._lab.results())[0]
        self.assertNotIn(es._mean('e'), rc[Experiment.RESULTS])
                         
    def testNoPoint( self ):
        '''Test we do nothing if we have an empty parameter space.'''
        e = SampleExperiment5()
        self._lab.runExperiment(e)
        self.assertEqual(len(self._lab.results()), 0)
        
    def testExtrema( self ):
        '''Test we capture the extrema correctly.'''
        N = 10
        self._lab['x'] = [ 5 ]    # dummy data point, we don't use it
        
        e = SampleExperiment5()
        es = SummaryExperiment(RepeatedExperiment(e, N))

        self._lab.runExperiment(es)
        self.assertTrue(es.success())
        res = (self._lab.results())[0]

        self.assertEqual(res[Experiment.METADATA][SummaryExperiment.UNDERLYING_SUCCESSFUL_RESULTS], 10)
        self.assertEqual(res[Experiment.RESULTS]['result_min'], numpy.min(e._values))
        self.assertEqual(res[Experiment.RESULTS]['result_max'], numpy.max(e._values))



# Tests of result sets
#
# Copyright (C) 2016--2020 Simon Dobson
# 
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import *

import unittest
from datetime import datetime
import numpy
import pandas


class SampleExperiment(Experiment):
    '''A very simple experiment that adds up its parameters.'''

    def do( self, param ):
        total = 0
        for k in param:
            total = total + param[k]
        return dict(total = total)


class ResultSetTests(unittest.TestCase):

    def setUp(self):
        '''Populate a results dict for testing with.'''
        self._rs = ResultSet()

        # populate metadata of otherwise empty results dict
        self._rc = Experiment.resultsdict()
        self._rc[Experiment.METADATA][Experiment.EXPERIMENT] = str(self.__class__)
        self._rc[Experiment.METADATA][Experiment.START_TIME] = datetime.now()
        self._rc[Experiment.METADATA][Experiment.END_TIME] = datetime.now()
        self._rc[Experiment.METADATA][Experiment.SETUP_TIME] = 10
        self._rc[Experiment.METADATA][Experiment.EXPERIMENT_TIME] = 20
        self._rc[Experiment.METADATA][Experiment.TEARDOWN_TIME] = 10
        self._rc[Experiment.METADATA][Experiment.ELAPSED_TIME] = 40
        self._rc[Experiment.METADATA][Experiment.STATUS] = True
        self._rc[Experiment.METADATA][Experiment.STATUS] = True
        self._rc[Experiment.PARAMETERS]['singleton'] = 1
        self._rc[Experiment.RESULTS]['first'] = 1

    def testTitle(self):
        '''Test we can set descriptions, and get default ones.'''
        rs1 = ResultSet('My results')
        self.assertEqual(rs1.description(), 'My results')
        rs2 = ResultSet()
        self.assertIsNotNone(rs2.description())


    # ---------- Type inference ----------

    def testNoDtype(self):
        '''Test we initially have no dtype set.'''''
        with self.assertRaises(Exception):
            self._rs.dtype()

    def testInitialInferSuccess(self):
        '''Test we can infer the initial dtype for a successful results dict.'''
        self._rc[Experiment.METADATA][Experiment.STATUS] = True
        self._rc[Experiment.PARAMETERS]['k'] = 1
        self._rc[Experiment.RESULTS]['total'] = 2.0
        self._rc[Experiment.RESULTS]['flag'] = False
        dtype = self._rs.inferDtype(self._rc)
        self.assertCountEqual(self._rs.metadataNames(), Experiment.StandardMetadata)
        self.assertCountEqual(self._rs.parameterNames(), ['k', 'singleton'])
        self.assertCountEqual(self._rs.resultNames(), ['total', 'first', 'flag'])
        self.assertEqual(dtype.fields['k'][0], numpy.dtype(ResultSet.TypeMapping[int]))
        self.assertEqual(dtype.fields['total'][0], numpy.dtype(ResultSet.TypeMapping[float]))
        self.assertEqual(dtype.fields['singleton'][0], numpy.dtype(ResultSet.TypeMapping[int]))
        self.assertEqual(dtype.fields['first'][0], numpy.dtype(ResultSet.TypeMapping[int]))
        self.assertEqual(dtype.fields['flag'][0], numpy.dtype(ResultSet.TypeMapping[bool]))
        self.assertEqual(dtype.fields[Experiment.START_TIME][0], numpy.dtype(ResultSet.TypeMapping[datetime]))
        self.assertEqual(dtype.fields[Experiment.EXCEPTION][0], numpy.dtype(ResultSet.TypeMapping[str]))

    def testInitialInferFailure(self):
        '''Test we can infer the initial dtype for an unsuccessful results dict.'''
        self._rc[Experiment.METADATA][Experiment.STATUS] = False
        self._rc[Experiment.METADATA][Experiment.EXCEPTION] = str(Exception('wrong'))
        self._rc[Experiment.METADATA][Experiment.TRACEBACK] = '<backtrace>'
        self._rc[Experiment.PARAMETERS]['k'] = 1
        dtype = self._rs.inferDtype(self._rc)
        self.assertCountEqual(self._rs.metadataNames(), Experiment.StandardMetadata)
        self.assertCountEqual(self._rs.parameterNames(), ['k', 'singleton'])
        self.assertCountEqual(self._rs.resultNames(), [])
        self.assertEqual(dtype.fields['k'][0], numpy.dtype(ResultSet.TypeMapping[int]))
        self.assertEqual(dtype.fields[Experiment.START_TIME][0], numpy.dtype(ResultSet.TypeMapping[datetime]))
        self.assertEqual(dtype.fields[Experiment.EXCEPTION][0], numpy.dtype(ResultSet.TypeMapping[str]))
        self.assertEqual(dtype.fields[Experiment.TRACEBACK][0], numpy.dtype(ResultSet.TypeMapping[str]))

    def testAddMetadata(self):
        '''Test we can add metadata elements beyond the standard set.'''
        self._rc[Experiment.METADATA]['additional'] = True
        self._rc[Experiment.PARAMETERS]['k'] = 1
        dtype = self._rs.inferDtype(self._rc)
        self.assertCountEqual(self._rs.metadataNames(), Experiment.StandardMetadata.union(set(['additional'])))
        self.assertEqual(dtype.fields['additional'][0], numpy.dtype(ResultSet.TypeMapping[bool]))

        # check we can now extend the
        #  metadata
        self._rc[Experiment.METADATA]['andagain'] = 'here we go'
        dtype = self._rs.inferDtype(self._rc)
        self.assertCountEqual(self._rs.metadataNames(), Experiment.StandardMetadata.union(set(['additional', 'andagain'])))
        self.assertEqual(dtype.fields['andagain'][0], numpy.dtype(ResultSet.TypeMapping[str]))

    def testAddResults(self):
        '''Test we can add more results.'''

        # initial set of paramneters and results
        self._rc[Experiment.METADATA][Experiment.STATUS] = True
        self._rc[Experiment.PARAMETERS]['k'] = 1
        self._rc[Experiment.RESULTS]['total'] = 2.0
        self._rc[Experiment.RESULTS]['flag'] = False
        self._rs.inferDtype(self._rc)
        self.assertCountEqual(self._rs.resultNames(), ['total', 'first', 'flag'])

        # add a result with an extra field and no flag, dtype should be extended
        self._rc[Experiment.PARAMETERS]['k'] = 2
        self._rc[Experiment.RESULTS]['total'] = 3.0
        del self._rc[Experiment.RESULTS]['flag']
        self._rc[Experiment.RESULTS]['extra'] = 'hello'
        dtype = self._rs.inferDtype(self._rc)
        self.assertCountEqual(self._rs.metadataNames(), Experiment.StandardMetadata)
        self.assertCountEqual(self._rs.parameterNames(), ['k', 'singleton'])
        self.assertCountEqual(self._rs.resultNames(), ['total', 'first', 'flag', 'extra'])
        self.assertEqual(dtype.fields['flag'][0], numpy.dtype(ResultSet.TypeMapping[bool]))
        self.assertEqual(dtype.fields['extra'][0], numpy.dtype(ResultSet.TypeMapping[str]))

    def testAddParametersAndResults(self):
        '''Test we can add more parameters and more results.'''

        # initial set of parameters and results
        self._rc[Experiment.METADATA][Experiment.STATUS] = True
        self._rc[Experiment.PARAMETERS]['k'] = 1
        self._rc[Experiment.RESULTS]['total'] = 2.0
        self._rc[Experiment.RESULTS]['flag'] = False
        self._rs.inferDtype(self._rc)
        self.assertCountEqual(self._rs.parameterNames(), ['k', 'singleton'])
        self.assertCountEqual(self._rs.resultNames(), ['total', 'first', 'flag'])
        self.assertFalse(self._rs.isDirty())
        self.assertTrue(self._rs.isTypeChanged())

        # add a result with extra parameters and results
        self._rs.typechanged(False)
        self._rc[Experiment.PARAMETERS]['k'] = 2
        self._rc[Experiment.PARAMETERS]['j'] = 9
        self._rc[Experiment.RESULTS]['total'] = 3.0
        self._rc[Experiment.RESULTS]['flag'] = False
        self._rc[Experiment.RESULTS]['extra'] = 'hello'
        dtype = self._rs.inferDtype(self._rc)
        self.assertCountEqual(self._rs.metadataNames(), Experiment.StandardMetadata)
        self.assertCountEqual(self._rs.parameterNames(), ['k', 'j', 'singleton'])
        self.assertCountEqual(self._rs.resultNames(), ['total', 'first', 'flag', 'extra'])
        self.assertEqual(dtype.fields['extra'][0], numpy.dtype(ResultSet.TypeMapping[str]))
        self.assertTrue(self._rs.isTypeChanged())

    def testAddInfer(self):
        '''Test that adding a result does the inference operation.'''
        self._rc[Experiment.METADATA][Experiment.STATUS] = True
        self._rc[Experiment.PARAMETERS]['k'] = 1
        self._rc[Experiment.RESULTS]['total'] = 2.0
        self._rc[Experiment.RESULTS]['flag'] = False
        self._rs.addSingleResult(self._rc)
        dtype = self._rs.dtype()
        self.assertCountEqual(self._rs.parameterNames(), ['k', 'singleton'])
        self.assertCountEqual(self._rs.resultNames(), ['total', 'first', 'flag'])
        self.assertEqual(dtype.fields['k'][0], numpy.dtype(ResultSet.TypeMapping[int]))
        self.assertEqual(dtype.fields['total'][0], numpy.dtype(ResultSet.TypeMapping[float]))
        self.assertEqual(dtype.fields['flag'][0], numpy.dtype(ResultSet.TypeMapping[bool]))
        self.assertTrue(self._rs.isDirty())

    def testExtendDataframe(self):
        '''Test that extending the dtype extends the dataframe properly.'''

        # add first result
        self._rc[Experiment.METADATA][Experiment.STATUS] = True
        self._rc[Experiment.PARAMETERS]['k'] = 1
        self._rc[Experiment.RESULTS]['total'] = 2.0
        self._rc[Experiment.RESULTS]['flag'] = False
        self._rs.addSingleResult(self._rc)

        # add result with extra result elements, dataframe should gain columns
        # with default values in the existing rows
        self._rc[Experiment.PARAMETERS]['k'] = 2
        self._rc[Experiment.PARAMETERS]['j'] = 9
        self._rc[Experiment.RESULTS]['total'] = 3.0
        self._rc[Experiment.RESULTS]['flag'] = False
        self._rc[Experiment.RESULTS]['extra'] = 'hello'
        self._rs.addSingleResult(self._rc)
        df = self._rs.dataframe()
        self.assertTrue((df[df['k'] == 1]['extra'] == '').all())
        self.assertTrue((df[df['k'] == 2]['extra'] == 'hello').all())
        self.assertTrue(self._rs.isDirty())

    def testInferPending(self):
        '''Test we infer the dtype correct for pending results.'''
        self._rc[Experiment.PARAMETERS]['k'] = 1
        dtype = self._rs.inferPendingResultDtype(self._rc[Experiment.PARAMETERS])
        self.assertCountEqual(self._rs.metadataNames(), [])
        self.assertCountEqual(self._rs.parameterNames(), ['k', 'singleton'])
        self.assertCountEqual(self._rs.resultNames(), [])
        self.assertEqual(dtype.fields['k'][0], numpy.dtype(ResultSet.TypeMapping[int]))
        self.assertEqual(dtype.fields['singleton'][0], numpy.dtype(ResultSet.TypeMapping[int]))

    def testAddPendingParameters(self):
        '''Test we can add parameters to pending results.'''

        # add initial pending result
        self._rc[Experiment.PARAMETERS]['k'] = 1
        dtype = self._rs.inferPendingResultDtype(self._rc[Experiment.PARAMETERS])
        self.assertCountEqual(self._rs.metadataNames(), [])
        self.assertCountEqual(self._rs.parameterNames(), ['k', 'singleton'])
        self.assertCountEqual(self._rs.resultNames(), [])
        self.assertEqual(dtype.fields['k'][0], numpy.dtype(ResultSet.TypeMapping[int]))

        # add another pending result with extended parameters, dtype should be extended
        self._rc[Experiment.PARAMETERS]['k'] = 2
        self._rc[Experiment.PARAMETERS]['j'] = 34.56
        dtype = self._rs.inferPendingResultDtype(self._rc[Experiment.PARAMETERS])
        self.assertCountEqual(self._rs.metadataNames(), [])
        self.assertCountEqual(self._rs.parameterNames(), ['k', 'j', 'singleton'])
        self.assertCountEqual(self._rs.resultNames(), [])
        self.assertEqual(dtype.fields['k'][0], numpy.dtype(ResultSet.TypeMapping[int]))
        self.assertEqual(dtype.fields['j'][0], numpy.dtype(ResultSet.TypeMapping[float]))
        self.assertEqual(dtype.fields['singleton'][0], numpy.dtype(ResultSet.TypeMapping[int]))

    def testInferResolve(self):
        '''Test the interaction between pending and real results when inferring type.'''

        # set initial pending dtype from parameters
        self._rc[Experiment.PARAMETERS]['k'] = 1
        self._rs.inferPendingResultDtype(self._rc[Experiment.PARAMETERS])

        # results, same parameters
        self._rc[Experiment.RESULTS]['total'] = 2.0
        dtype = self._rs.inferDtype(self._rc)
        self.assertCountEqual(self._rs.metadataNames(), Experiment.StandardMetadata)
        self.assertCountEqual(self._rs.parameterNames(), ['k', 'singleton'])
        self.assertCountEqual(self._rs.resultNames(), ['total', 'first'])
        self.assertEqual(dtype.fields['k'][0], numpy.dtype(ResultSet.TypeMapping[int]))
        self.assertEqual(dtype.fields['total'][0], numpy.dtype(ResultSet.TypeMapping[float]))

        # more results, same types, shouldn't change the dtype
        self._rs.dirty(False)
        self._rc[Experiment.PARAMETERS]['k'] = 2
        self._rc[Experiment.RESULTS]['total'] = 3.0
        dtype2 = self._rs.inferDtype(self._rc)
        self.assertEqual(dtype, dtype2)
        self.assertFalse(self._rs.isDirty())

    def testInferResolveExtend(self):
        '''Test we can extend parameters in a real (resolving) result.'''

        # set initial pending dtype
        self._rc[Experiment.PARAMETERS]['k'] = 1
        dtype = self._rs.inferPendingResultDtype(self._rc[Experiment.PARAMETERS])

        # results, same types, extra parameters
        self._rc[Experiment.PARAMETERS]['j'] = 34.56
        self._rc[Experiment.RESULTS]['total'] = 2.0
        dtype = self._rs.inferDtype(self._rc)
        self.assertCountEqual(self._rs.metadataNames(), Experiment.StandardMetadata)
        self.assertCountEqual(self._rs.parameterNames(), ['k', 'j', 'singleton'])
        self.assertCountEqual(self._rs.resultNames(), ['total', 'first'])
        self.assertEqual(dtype.fields['k'][0], numpy.dtype(ResultSet.TypeMapping[int]))
        self.assertEqual(dtype.fields['j'][0], numpy.dtype(ResultSet.TypeMapping[float]))
        self.assertEqual(dtype.fields['total'][0], numpy.dtype(ResultSet.TypeMapping[float]))


    # ---------- Attributes ----------

    def testAttributes(self):
        '''Test all the attribute operators.'''''
        
        # set and get
        self._rs['number1'] = '1'
        self.assertEqual(self._rs['number1'], '1')
        self.assertEqual(len(self._rs.keys()), 1)
        self.assertEqual(len(self._rs), 0)              # len() refers to results

        # check contains
        self.assertIn('number1', self._rs)
        self.assertNotIn('number2', self._rs)

        # exceptions for undefined attributes
        with self.assertRaises(Exception):
            self._rs['result2']
 
         # check we can delete attributes
        del self._rs['number1']
        with self.assertRaises(Exception):
            self._rs['result1']
        self.assertEqual(len(self._rs.keys()), 0)
        self.assertNotIn('number1', self._rs)


    # ---------- Adding results ----------

    def testaddSingleResult(self):
        '''Test we can add results to the set.'''
        self.assertCountEqual(self._rs.parameterNames(), [])

        # test initial result sets parameter names
        self._rs.addSingleResult(self._rc)
        self.assertCountEqual(self._rc[Experiment.PARAMETERS].keys(), self._rs.parameterNames())

        # add another result with different values
        self._rc[Experiment.PARAMETERS]['singleton'] = 2
        self._rs.addSingleResult(self._rc)

        # add a result with an extra parameter, should extend the dataframe
        self._rc[Experiment.PARAMETERS]['singleton'] = 3
        self._rc[Experiment.PARAMETERS]['radically'] = 'wrong'
        self._rs.addSingleResult(self._rc)
        df = self._rs.dataframe()
        self.assertTrue((df[df['singleton'] == 3]['radically'] == 'wrong').all())
        self.assertTrue((df[df['singleton'] != 3]['radically'] == '').all())
 
    def testSingleResult(self):
        '''Test retrieval of a single result.'''
        self._rs.addSingleResult(self._rc)
        df = self._rs.dataframeFor(self._rc[Experiment.PARAMETERS])
        self.assertEqual(len(df.index), 1)

        # add another result with same parameters, check we get them both
        self._rc[Experiment.RESULTS]['first'] = 2
        self._rs.addSingleResult(self._rc)
        df = self._rs.dataframeFor(self._rc[Experiment.PARAMETERS])
        self.assertEqual(len(df.index), 2)

        # check we don't get results when there are none
        self._rc[Experiment.PARAMETERS]['singleton'] = 5
        df = self._rs.dataframeFor(self._rc[Experiment.PARAMETERS])
        self.assertEqual(len(df.index), 0)

    def testSingleResultExtended(self):
        '''Test we can retrieve a result when the dataframe was extended with another.'''
        # add initial result
        self._rs.addSingleResult(self._rc)
        params1 = self._rc[Experiment.PARAMETERS].copy()

        # add a result with an extra parameter
        self._rc[Experiment.PARAMETERS]['singleton'] = 3
        self._rc[Experiment.PARAMETERS]['radically'] = 'wrong'
        self._rs.addSingleResult(self._rc)

        # retrieve the first result without knowing about the extended parameters
        df = self._rs.dataframeFor(params1)
        self.assertEqual(len(df.index), 1)
        self.assertTrue((df[df['singleton'] == params1['singleton']]['first'] == 1).all())
        self.assertTrue((df[df['singleton'] == params1['singleton']]['radically'] == '').all())
        self.assertTrue((df[df['singleton'] == 3]['first'] == 1).all())
        self.assertTrue((df[df['singleton'] == 3]['radically'] == 'wrong').all())

    def testSmallerParameters(self):
        '''Test we can project out values properly with fewer parameters provided.'''

        # add some results
        self._rc[Experiment.PARAMETERS]['singleton'] = 1
        self._rc[Experiment.PARAMETERS]['additional'] = 0
        self._rs.addSingleResult(self._rc)
        self._rc[Experiment.PARAMETERS]['singleton'] = 5
        self._rc[Experiment.RESULTS]['first'] = 2
        self._rs.addSingleResult(self._rc)
        self._rc[Experiment.PARAMETERS]['additional'] = 1
        self._rc[Experiment.RESULTS]['first'] = 2
        self._rs.addSingleResult(self._rc)

        # first sub-set, extracts several results
        ps = dict()
        ps['additional'] = 0
        df = self._rs.dataframeFor(ps)
        self.assertEqual(len(df.index), 2)

        # second sub-set, extracts one result
        ps['additional'] = 1
        df = self._rs.dataframeFor(ps)
        self.assertEqual(len(df.index), 1)

    def testConjunction(self):
        '''Test we can project-out results.'''

        # add some results
        self._rc[Experiment.PARAMETERS]['singleton'] = 1
        self._rc[Experiment.PARAMETERS]['additional'] = 0
        self._rs.addSingleResult(self._rc)
        self._rc[Experiment.PARAMETERS]['singleton'] = 5
        self._rc[Experiment.RESULTS]['first'] = 2
        self._rs.addSingleResult(self._rc)
        self._rc[Experiment.PARAMETERS]['additional'] = 1
        self._rc[Experiment.RESULTS]['first'] = 2
        self._rs.addSingleResult(self._rc)

        # progressively refine query
        ps = dict()
        ps['additional'] = 0
        df = self._rs.dataframeFor(ps)
        self.assertEqual(len(df.index), 2)
        ps['singleton'] = 5
        df = self._rs.dataframeFor(ps)
        self.assertEqual(len(df.index), 1)
        ps['singleton'] = 0
        df = self._rs.dataframeFor(ps)
        self.assertEqual(len(df.index), 0)

    def testMultiple(self):
        '''Test we can project out multiple values of the same parameter.'''
        self._rc[Experiment.PARAMETERS]['singleton'] = 1
        self._rc[Experiment.PARAMETERS]['additional'] = 0
        self._rs.addSingleResult(self._rc)
        self._rc[Experiment.PARAMETERS]['singleton'] = 5
        self._rc[Experiment.RESULTS]['first'] = 2
        self._rs.addSingleResult(self._rc)
        self._rc[Experiment.PARAMETERS]['additional'] = 1
        self._rc[Experiment.RESULTS]['first'] = 2
        self._rs.addSingleResult(self._rc)

        # extract all results with parameter in range
        ps = dict()
        ps['additional'] = [0, 1]
        df = self._rs.dataframeFor(ps)
        self.assertEqual(len(df.index), 3)

        # refine query to exclude some other results
        ps['singleton'] = 5
        df = self._rs.dataframeFor(ps)
        self.assertEqual(len(df.index), 2)

    def testImmutable(self):
        '''Test we can't update the results we get back.'''
        self._rs.addSingleResult(self._rc)
        df = self._rs.dataframeFor(self._rc[Experiment.PARAMETERS])

        # update-in-place on the dataframe we were passed
        v = self._rc[Experiment.RESULTS]['first']
        nv = v + 5
        df.loc[0, 'first'] = nv
        self.assertTrue((df['first'] == nv).all())

        # make sure we didn't affect the result set
        df = self._rs.dataframeFor(self._rc[Experiment.PARAMETERS])
        self.assertTrue((df['first'] == v).all())

    def testEmptyParameters(self):
        '''Test we get all results if we don't do any projection.'''
        self._rs.addSingleResult(self._rc)
        df = self._rs.dataframeFor(dict())
        self.assertEqual(len(df.index), 1)

    def testNumberOfResultsZero(self):
        '''Test we can handle zero results.'''
        self.assertEqual(self._rs.numberOfResults(), 0)        

    def testNumberOfResultsOne(self):
        '''Test we can count results.'''
        self._rs.addSingleResult(self._rc)
        self.assertEqual(self._rs.numberOfResults(), 1)        

    def testAddNoResults(self):
        '''Test we can correctly add the results of an experiment that doesn't have any actual results.'''
        self._rc[Experiment.RESULTS] = dict()
        self._rs.addSingleResult(self._rc)
        self.assertEqual(self._rs.numberOfResults(), 1)        
        self.assertCountEqual(self._rs.resultNames(), [])
        ress = self._rs.resultsFor(self._rc[Experiment.PARAMETERS])
        self.assertEqual(len(ress), 1)
        res = ress[0]
        self.assertIn(Experiment.RESULTS, res)
        self.assertDictEqual(res[Experiment.RESULTS], dict())        

    def testInferList(self):
        '''Test we can infer a list type for a result.'''

        # add a result with a list
        self._rc[Experiment.PARAMETERS]['singleton'] = 1
        self._rc[Experiment.RESULTS]['list'] = [ 1, 2, 3 ]
        self._rs.addSingleResult(self._rc)
        
        # add another with the same shape1
        self._rc[Experiment.PARAMETERS]['singleton'] = 2
        self._rc[Experiment.RESULTS]['list'] = [ 4, 5, 6 ]
        self._rs.addSingleResult(self._rc)

        # add another with a different shape
        self._rc[Experiment.PARAMETERS]['singleton'] = 3
        self._rc[Experiment.RESULTS]['list'] = [ 14, 15, 16, 17, 18 ]
        self._rs.addSingleResult(self._rc)

        # check the values have the right shapes
        params = dict()
        params['singleton'] = 1
        rc = self._rs.resultsFor(params)[0]
        self.assertCountEqual(rc[Experiment.RESULTS]['list'], [ 1, 2, 3 ])


    # ---------- Pending results ----------

    def testPendingResults(self):
        '''Test we can add and resolve pending results.'''
        self._rs.addSinglePendingResult(self._rc[Experiment.PARAMETERS], '1234')
        self.assertCountEqual(self._rs.pendingResults(), [ '1234' ])

        self._rs.addSingleResult(self._rc)
        self._rs.cancelSinglePendingResult('1234')
        self.assertCountEqual(self._rs.pendingResults(), [])

        # check we have the cancelled result still
        rss = self._rs.dataframeFor(self._rc[Experiment.PARAMETERS])
        self.assertEqual(len(rss), 2)

        # check there's only one successful result
        rss = self._rs.dataframeFor(self._rc[Experiment.PARAMETERS], only_successful=True)
        self.assertEqual(len(rss), 1)
        self.assertEqual(rss['first'][0], 1)

    def testPendingResultsFor(self):
        '''Test we can retrieve pending results for a sub-set of parameters.'''
        params = dict()
        params['a'] = 10
        params['b'] = 50
        params['c'] = 'fifty'
        self.assertEqual(len(self._rs.pendingResultsFor(params)), 0)
        self._rs.addSinglePendingResult(params, '1234')
        self.assertEqual(len(self._rs.pendingResultsFor(params)), 1)
        params['b'] = 90
        params['c'] = 'ninety'
        self._rs.addSinglePendingResult(params, '5678')
        params['b'] = 90
        params['c'] = 'ninety'
        self._rs.addSinglePendingResult(params, '91011')
        self.assertEqual(len(self._rs.pendingResultsFor(dict(a=10))), 3)
        self.assertEqual(len(self._rs.pendingResultsFor(dict(b=50))), 1)
        self.assertEqual(len(self._rs.pendingResultsFor(dict(a=10, b=50))), 1)
        self.assertEqual(len(self._rs.pendingResultsFor(dict(a=15, b=50))), 0)

    def testNumberOfPendingResultsZero(self):
        '''Test we can handle zero pending results.'''
        self.assertEqual(self._rs.numberOfPendingResults(), 0)        

    def testNumberOfPendingResultsOne(self):
        '''Test we can count pending results.'''
        self._rs.addSinglePendingResult(self._rc[Experiment.PARAMETERS], '1234')
        self.assertEqual(self._rs.numberOfPendingResults(), 1)        

    def testJobId(self):
        '''Test we raise exceptions for unrecognised job ids.'''
        self._rs.addSinglePendingResult(self._rc[Experiment.PARAMETERS], '1234')
        with self.assertRaises(PendingResultException):
            self._rs.cancelSinglePendingResult('4567')
        with self.assertRaises(PendingResultException):
            self._rs.resolveSinglePendingResult('4567')

    def testDuplicateJobs(self):
        '''Test we catch addition of duplicate job ids.'''
        self._rs.addSinglePendingResult(self._rc[Experiment.PARAMETERS], '1234')
        self._rc[Experiment.PARAMETERS]['singleton'] = 3
        with self.assertRaises(Exception):
            self._rs.addSinglePendingResult(self._rc[Experiment.PARAMETERS], '1234')

    def testCancelNonExistentJob(self):
        '''Test we can't cancel non-existent job ids.'''
        self._rs.addSinglePendingResult(self._rc[Experiment.PARAMETERS], '1234')
        with self.assertRaises(Exception):
            self._rs.cancelSinglePendingResult('4567')

    def testCancelOnce(self):
        '''Test that a job can't be re-canceled.'''
        self._rs.addSinglePendingResult(self._rc[Experiment.PARAMETERS], '1234')
        self._rs.cancelSinglePendingResult('1234')
        with self.assertRaises(Exception):
            self._rs.cancelSinglePendingResult('1234')

    def testDuplicatePendingParameters(self):
        '''Test we can have multiple pending jobs with the same parameters.'''
        self._rs.addSinglePendingResult(self._rc[Experiment.PARAMETERS], '1234')
        self._rs.addSinglePendingResult(self._rc[Experiment.PARAMETERS], '4567')
        self.assertCountEqual(self._rs.pendingResults(), [ '1234', '4567' ])

        self._rc[Experiment.RESULTS]['first'] = 1
        self._rs.addSingleResult(self._rc)
        self._rs.cancelSinglePendingResult('1234')
        self._rc[Experiment.RESULTS]['first'] = 2
        self._rs.addSingleResult(self._rc)
        self._rs.cancelSinglePendingResult('4567')
        self.assertCountEqual(self._rs.pendingResults(), [])

        # check the two cancelled results were zeroed correctly
        rss = self._rs.dataframeFor(self._rc[Experiment.PARAMETERS])
        self.assertEqual(len(rss), 4)
        self.assertCountEqual(rss['first'], [ 0, 0, 1, 2 ])

        # check that we project out the right successes
        rsss = rss[rss[Experiment.STATUS] == True]
        self.assertEqual(len(rsss), 2)
        self.assertCountEqual(rsss['first'], [ 1, 2 ])

    def testPendingParameterNames(self):
        '''Test we have to respect parameter names when adding pending jobs.'''
        self._rc[Experiment.PARAMETERS]['duplicate'] = 10
        self._rs.addSingleResult(self._rc)

        # too many parameters, should be fine
        self._rc[Experiment.PARAMETERS]['triplicate'] = 20
        self._rs.addSinglePendingResult(self._rc[Experiment.PARAMETERS], '1234')

        # too few parameters, should fail
        del self._rc[Experiment.PARAMETERS]['triplicate']
        del self._rc[Experiment.PARAMETERS]['duplicate']
        with self.assertRaises(Exception):   
            self._rs.addSinglePendingResult(self._rc[Experiment.PARAMETERS], '1234')

    def testCancelJobById(self):
        '''Test we can cancel jobs by their identifier.'''
        self._rc[Experiment.PARAMETERS]['duplicate'] = 10
        self._rs.addSinglePendingResult(self._rc[Experiment.PARAMETERS], '1234')
        self._rc[Experiment.PARAMETERS]['duplicate'] = 20
        self._rs.addSinglePendingResult(self._rc[Experiment.PARAMETERS], '4567')
        self.assertCountEqual(self._rs.pendingResults(), [ '1234', '4567' ])

        # check job gets cancelled
        self._rs.cancelSinglePendingResult('1234')
        self.assertCountEqual(self._rs.pendingResults(), [ '4567' ])

        # check the exception is correct
        rcs = self._rs.resultsFor(dict(duplicate=10))
        self.assertEqual(len(rcs), 1)
        rc = rcs[0]
        self.assertTrue(isinstance(rc[Experiment.METADATA][Experiment.EXCEPTION], CancelledException))

    def testFailedResult(self):
        '''Test we can save a failed result, i.e., with an exception and no results.'''
        self._rc[Experiment.METADATA][Experiment.STATUS] = False
        self._rc[Experiment.METADATA][Experiment.EXCEPTION] = "An exception"
        self._rc[Experiment.METADATA][Experiment.TRACEBACK] = "...happened somewhere..."
        del self._rc[Experiment.RESULTS]
        self._rs.addSingleResult(self._rc)
        self.assertEqual(len(self._rs), 1)
        rcs = self._rs.resultsFor(self._rc[Experiment.PARAMETERS])
        self.assertEqual(len(rcs), 1)
        rc = rcs[0]
        self.assertDictEqual(self._rc[Experiment.METADATA], rc[Experiment.METADATA])
        self.assertDictEqual(self._rc[Experiment.PARAMETERS], rc[Experiment.PARAMETERS])

    def _resultsEqual(self, df, rc):
        '''Check that the dataframe contains a row with the given results.

        :param df: the dataframe
        :param rc: the results dict
        :returns: True if there's a corresponding row'''
        for d in [ Experiment.PARAMETERS, Experiment.RESULTS ]:  # ignore metadata
            for k in rc[d].keys():
                df = df[df[k] == rc[d][k]]
        return (len(df) > 0)

    def testRealAndPendingResults( self ):
        '''Test a sequence of real and pending results'''
        e = SampleExperiment()
        
        # create three results
        params1 = dict(a  = 1, b = 2)
        rc1 = e.set(params1).run()
        params2 = dict(a  = 10, b = 12)
        rc2 = e.set(params2).run()
        params3 = dict(a  = 45, b = 11)
        rc3 = e.set(params3).run()

        # add first result
        self._rs.addSingleResult(rc1)
        self.assertTrue(self._resultsEqual(self._rs.dataframeFor(params1), rc1))
        self.assertEqual(len(self._rs.dataframeFor(params2)), 0)
        self.assertEqual(len(self._rs.results()), 1)
        self.assertCountEqual(self._rs.pendingResults(), [])
        self.assertTrue(self._rs.ready())
        
        # add second result as pending
        self._rs.addSinglePendingResult(params2, '2')
        self.assertTrue(self._resultsEqual(self._rs.dataframeFor(params1), rc1))
        self.assertEqual(len(self._rs.dataframeFor(params2)), 0)
        self.assertEqual(len(self._rs.results()), 1)
        self.assertCountEqual(self._rs.pendingResults(), [ '2' ])
        self.assertFalse(self._rs.ready())
        
        # add third result as pending
        self._rs.addSinglePendingResult(params3, '3')
        self.assertTrue(self._resultsEqual(self._rs.dataframeFor(params1), rc1))
        self.assertEqual(len(self._rs.dataframeFor(params2)), 0)
        self.assertEqual(len(self._rs.dataframeFor(params3)), 0)
        self.assertEqual(len(self._rs.results()), 1)
        self.assertCountEqual(self._rs.pendingResults(), [ '2', '3' ])
        self.assertFalse(self._rs.ready())

        # resolve second result
        self._rs.addSingleResult(rc2)
        self._rs.cancelSinglePendingResult('2')
        self.assertTrue(self._resultsEqual(self._rs.dataframeFor(params1), rc1))
        self.assertTrue(self._resultsEqual(self._rs.dataframeFor(params2), rc2))
        self.assertEqual(len(self._rs.dataframeFor(params3)), 0)
        self.assertEqual(len(self._rs.results()), 3)
        self.assertCountEqual(self._rs.pendingResults(), [ '3' ])
        self.assertFalse(self._rs.ready())

        # cancel third result
        self._rs.cancelSinglePendingResult('3')
        self.assertTrue(self._resultsEqual(self._rs.dataframeFor(params1), rc1))
        self.assertTrue(self._resultsEqual(self._rs.dataframeFor(params2), rc2))
        self.assertEqual(len(self._rs.dataframeFor(params3)), 1)
        self.assertEqual(len(self._rs.dataframeFor(params3, only_successful=True)), 0)
        self.assertEqual(len(self._rs.results()), 4)
        self.assertCountEqual(self._rs.pendingResults(), [])
        self.assertTrue(self._rs.ready())

        # add pending result again under the same job id
        self._rs.addSinglePendingResult(params3, '3')
        self.assertTrue(self._resultsEqual(self._rs.dataframeFor(params1), rc1))
        self.assertTrue(self._resultsEqual(self._rs.dataframeFor(params2), rc2))
        self.assertEqual(len(self._rs.dataframeFor(params3)), 1)
        self.assertEqual(len(self._rs.results()), 4)
        self.assertCountEqual(self._rs.pendingResults(), [ '3' ])
        self.assertFalse(self._rs.ready())

        # cancel it again
        self._rs.cancelSinglePendingResult('3')
        self.assertTrue(self._resultsEqual(self._rs.dataframeFor(params1), rc1))
        self.assertTrue(self._resultsEqual(self._rs.dataframeFor(params2), rc2))
        self.assertEqual(len(self._rs.dataframeFor(params3)), 2)
        self.assertEqual(len(self._rs.results()), 5)
        self.assertCountEqual(self._rs.pendingResults(), [])
        self.assertTrue(self._rs.ready())

    def testLocking(self):
        '''Test we can lock result sets.'''
        e = SampleExperiment()
        
        # create three results
        params1 = dict(a  = 1, b = 2)
        rc1 = e.set(params1).run()
        params2 = dict(a  = 10, b = 12)
        rc2 = e.set(params2).run()
        params3 = dict(a  = 45, b = 11)
        rc3 = e.set(params3).run()

        # add first result, with second pending
        self.assertFalse(self._rs.isLocked())
        self._rs.addSingleResult(rc1)
        self._rs.addSinglePendingResult(params2, '1234')

        # finish and lock the result set
        self._rs.finish()
        self.assertEqual(self._rs.numberOfResults(), 2)
        self.assertEqual(self._rs.numberOfPendingResults(), 0)
        self.assertTrue(self._rs.isLocked())

        # make sure we can't add further jobs or otherwise change the result set
        with self.assertRaises(ResultSetLockedException):
            self._rs.addSingleResult(rc3)
        with self.assertRaises(ResultSetLockedException):
            self._rs.addSinglePendingResult(rc3, '3456')
        with self.assertRaises(ResultSetLockedException):
            self._rs['attribute'] = 'not saved'

        # make sure we still have access
        self.assertEqual(len(self._rs.results()), 2)


if __name__ == '__main__':
    unittest.main()








# Tests of repeated experiments combinator
#
# Copyright (C) 2016--2021 Simon Dobson
# 
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import *

import unittest


class SampleExperiment1(Experiment):
        
    def do( self, params ):
        return dict(result = params['x'])

    
class RepeatedExperimentTests(unittest.TestCase):

    def setUp( self ):
        self._lab = Lab()
        
    def testRepetitionsOnePoint( self ):
        '''Test we can repeat an experiment at a single point'''
        N = 10
        
        self._lab['x'] = [ 5 ]
        
        e = SampleExperiment1()
        er = RepeatedExperiment(e, N)

        self._lab.runExperiment(er)
        self.assertTrue(er.success())

        df = self._lab.dataframe()
        self.assertEqual(len(df), N * len(self._lab['x']))
        self.assertTrue(df['result'].eq(self._lab['x'][0]).all())
        self.assertTrue(df[RepeatedExperiment.REPETITIONS].eq(N).all())
        self.assertCountEqual(df[RepeatedExperiment.I].values, range(N))
        
    def testRepetitionsMultiplePoint( self ):
        '''Test we can repeat an experiment across a parameter space'''
        N = 10
        
        self._lab['x'] = [ 5, 10, 15 ]
        
        e = SampleExperiment1()
        er = RepeatedExperiment(e, N)

        self._lab.runExperiment(er)
        self.assertTrue(er.success())

        df = self._lab.dataframe()
        self.assertEqual(len(df), N * len(self._lab['x']))

        for x in self._lab['x']:
            dfx = df[df['x'] == x]
            self.assertEqual(len(dfx), N)
            self.assertTrue(dfx['result'].eq(x).all())
            self.assertTrue(dfx[RepeatedExperiment.REPETITIONS].eq(N).all())
            self.assertCountEqual(dfx[RepeatedExperiment.I].values, range(N))

    # TODO: check nesting for repeated repetitions

if __name__ == '__main__':
    unittest.main()





# Tests of experiments class
#
# Copyright (C) 2016--2021 Simon Dobson
# 
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import *

import unittest
import time


class SampleExperiment0(Experiment):
    '''A base experiment that records the phases of the experiment.'''
    
    def __init__( self ):
        super(SampleExperiment0, self).__init__()
        self._ps = []
        
    def setUp( self, params ):
        self._ps.append('setup')

    def tearDown( self ):
        self._ps.append('teardown')
        
    def do( self, params ):
        self._ps.append('do')
        return dict()

    def report( self, params, meta, res ):
        ext = Experiment.report(self, params, meta, res)
        ext[self.METADATA]['phases'] = self._ps
        return ext

class SampleExperiment1(Experiment):
    '''An experiment that does literally nothing.'''
    
    def do( self, params ):
        pass

class SampleExperiment2(SampleExperiment0):
    '''An experiment that does nothing for 1s and returns a results dict.'''
    
    def do( self, params ):
        time.sleep(1)
        return SampleExperiment0.do(self, params)

class SampleExperiment3(SampleExperiment0):
    '''An experiment that does a calculation.'''
    
    def do( self, params ):
        return dict(result = params['a'] + params['b'])

class SampleExperiment4(SampleExperiment0):
    '''An experiment that makes sure there are timings to test.'''
    
    def setUp( self, params ):
        time.sleep(1)

    def tearDown( self ):
        time.sleep(1)

    def do( self, params ):
        time.sleep(1)
        return dict()

class SampleExperiment5(SampleExperiment0):
    '''An experiment that fails in its main action.'''
    
    def do( self, params ):
        raise Exception('We failed (on purpose)')
   
class SampleExperiment6(SampleExperiment0):
    '''An experiment that fails in its setup, and so should not do a teardown.'''
    
    def setUp( self, params ):
        raise Exception('We failed (on purpose)')

class SampleExperiment7(SampleExperiment0):
    '''An experiment that fails in its teardown.'''
    
    def tearDown( self ):
        raise Exception('We failed (on purpose)')

class SampleExperiment8(Experiment):
    '''An experiment that returns a lot of results dicts.'''

    def do(self, params):
        res = []
        rc = self.resultsdict()
        rc[Experiment.RESULTS] = dict(a=1, c=params['c'])
        res.append(rc)
        rc = self.resultsdict()
        rc[Experiment.RESULTS] = dict(a=2, c=params['c'])
        res.append(rc)
        return res

class ExperimentTests(unittest.TestCase):

    def testNoResults( self ):
        '''Test do() not returning results.'''
        e = SampleExperiment1()
        e.set(dict())
        res = e.run()
        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])

    def testPhases( self ):
        '''Test that phases execute correctly.'''
        e = SampleExperiment2()
        e.set(dict())
        res = e.run()
        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])
        self.assertIn('phases', res[Experiment.METADATA].keys())
        self.assertEqual(res[Experiment.METADATA]['phases'], [ 'setup', 'do', 'teardown' ])

    def testParameters( self ):
        '''Test that parameters are recorded properly.'''
        e = SampleExperiment2()
        params = dict(a = 1, b = 1.0, c = 'hello world')
        e.set(params)
        res = e.run()
        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])
        self.assertIn(Experiment.PARAMETERS, res.keys())
        for k in params.keys():
            self.assertEqual(res[Experiment.PARAMETERS][k], params[k])

    def testReporting( self ):
        '''Test that results are reported properly.'''
        e = SampleExperiment3()
        params = dict(a = 1, b = 2, c = 'hello world')
        e.set(params)
        res = e.run()
        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])
        self.assertIn('result', res[Experiment.RESULTS].keys())
        self.assertEqual(res[Experiment.RESULTS]['result'], params['a'] + params['b'])

    def testTiming( self ):
        '''Test that timings are plausible.'''
        e = SampleExperiment4()
        e.set(dict())
        res = e.run()
        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])

        timing = res[Experiment.METADATA]
        #print(timing)
        self.assertTrue(timing[Experiment.END_TIME] > timing[Experiment.START_TIME])
        self.assertTrue(timing[Experiment.ELAPSED_TIME] > 0)
        self.assertTrue(timing[Experiment.SETUP_TIME] > 0)
        self.assertTrue(timing[Experiment.EXPERIMENT_TIME] > 0)
        self.assertTrue(timing[Experiment.TEARDOWN_TIME] > 0)
        self.assertAlmostEqual(timing[Experiment.ELAPSED_TIME], timing[Experiment.SETUP_TIME] + timing[Experiment.TEARDOWN_TIME] + timing[Experiment.EXPERIMENT_TIME], places=3)

    def testException( self ):
        '''Test that exceptions are caught and reported in-line.'''
        e = SampleExperiment5()
        e.set(dict())
        res = e.run()
        self.assertFalse(res[Experiment.METADATA][Experiment.STATUS])
        self.assertIn('phases', res[Experiment.METADATA].keys())
        self.assertEqual(res[Experiment.METADATA]['phases'], [ 'setup', 'teardown' ])

    def testExceptionInSetup( self ):
        '''Test that exceptions in setup are caught and not torn down.'''
        e = SampleExperiment6()
        e.set(dict())
        res = e.run()
        self.assertFalse(res[Experiment.METADATA][Experiment.STATUS])
        self.assertIn('phases', res[Experiment.METADATA].keys())
        self.assertEqual(res[Experiment.METADATA]['phases'], [ ])

    def testExceptionInTeardown( self ):
        '''Test that exceptions in teardown are caught.'''
        e = SampleExperiment7()
        e.set(dict())
        res = e.run()
        self.assertFalse(res[Experiment.METADATA][Experiment.STATUS])
        self.assertIn('phases', res[Experiment.METADATA].keys())
        self.assertEqual(res[Experiment.METADATA]['phases'], [ 'setup', 'do' ])

    def testRecordingOnObject( self ):
        '''Test that everything is also available through the experiment object.'''
        e = SampleExperiment3()
        params = dict(a = 1, b = 2, c = 'hello world')
        e.set(params)
        res = e.run()
        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])
        self.assertEqual(res[Experiment.METADATA][Experiment.STATUS], e.success())
        self.assertIn('result', res[Experiment.RESULTS].keys())
        self.assertEqual(res[Experiment.RESULTS]['result'], params['a'] + params['b'])

    def testKeyInterface( self ):
        '''Test access to results by key.'''
        e = SampleExperiment3()
        params = dict(a = 1, b = 2, c = 'hello world')
        e.set(params)
        res = e.run()
        self.assertEqual(e['result'], params['a'] + params['b'])

    def testListResults(self):
        '''Test we wrap a list of experimental results properly.'''
        e = SampleExperiment8()
        params = dict(c='hello world')
        e.set(params)
        rc = e.run()
        self.assertFalse(isinstance(rc, list))
        self.assertTrue(isinstance(rc[Experiment.RESULTS], list))
        res = rc[Experiment.RESULTS]
        self.assertEqual(len(res), 2)
        for rc in res:
            self.assertEqual(rc[Experiment.RESULTS]['c'], params['c'])
       
if __name__ == '__main__':
    unittest.main()



# Tests of sequential lab class
#
# Copyright (C) 2016--2020 Simon Dobson
# 
# This file is part of epyc, experiment management in Python.
#
# epyc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# epyc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with epyc. If not, see <http://www.gnu.org/licenses/gpl.html>.

from epyc import *

import unittest
import numpy


class SampleExperiment(Experiment):
    '''A very simple experiment that adds up its parameters.'''

    def do( self, param ):
        total = 0
        for k in param:
            total = total + param[k]
        return dict(total = total)

    
class NullExperiment(Experiment):
    '''An experiment that just returns the value of its (one)
    parameter as its result.'''

    def do( self, param ):
        k = list(param.keys())[0]
        return dict(result = param[k])
    
        
class LabTests(unittest.TestCase):

    def setUp( self ):
        '''Create a lab in which to perform tests.'''
        self._lab = Lab()
        
    def testParameterOne( self ):
        '''Test that we can set a single parameter.'''
        self._lab['a'] = numpy.arange(0, 100)
        
        self.assertEqual(len(self._lab['a']), len(numpy.arange(0, 100)))
        for i in numpy.arange(0, 100):
            self.assertIn(i, self._lab['a'])

    def testParameterThree( self ):
        '''Test that we can set several parameters.'''
        self._lab['a'] = numpy.arange(0, 100)
        self._lab['b'] = numpy.arange(0, 1000)
        self._lab['c'] = numpy.arange(10, 50, 10)
        
        self.assertCountEqual(self._lab['a'], numpy.arange(0, 100))
        self.assertCountEqual(self._lab['b'], numpy.arange(0, 1000))
        self.assertCountEqual(self._lab['c'], numpy.arange(10, 50, 10))

    def testRunOne( self ):
        '''Test that a simple experiment runs against a 1D parameter space.'''
        n = 100

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab.runExperiment(SampleExperiment())
        res = self._lab.results()

        # check that the whole parameter space has a result
        self.assertEqual(len(res), n)
        for p in res:
            self.assertIn(p[Experiment.PARAMETERS]['a'], r)

        # check that each result corresponds to its parameter
        for p in res:
            self.assertEqual(p[Experiment.PARAMETERS]['a'], p[Experiment.RESULTS]['total'])

    def testRunTwo( self ):
        '''Test that a simple experiment runs against a 2D parameter space.'''
        n = 10

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab['b'] = r
        self._lab.runExperiment(SampleExperiment())
        res = self._lab.results()

        # check that the whole parameter space has a result
        self.assertEqual(len(res), n * n)
        for p in res:
            self.assertIn(p[Experiment.PARAMETERS]['a'], r)
            self.assertIn(p[Experiment.PARAMETERS]['b'], r)

        # check that each result corresponds to its parameter
        for p in res:
            self.assertEqual(p[Experiment.PARAMETERS]['a'] + p[Experiment.PARAMETERS]['b'], p[Experiment.RESULTS]['total'])
 
    def testReady(self):
        '''Test we can check readiness correctly.'''
        n = 10

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab['b'] = r
        self._lab.runExperiment(SampleExperiment())
        self.assertTrue(self._lab.ready())
        self.assertEqual(self._lab.readyFraction(), 1.0)

    def testSinglePoint( self ):
        '''Test that using a single point as a range still works.'''
        n = 100

        r = numpy.arange(0, n)
        self._lab['a'] = r
        self._lab['b'] = 0
        self._lab.runExperiment(SampleExperiment())
        res = self._lab.results()

        # check that the whole parameter space has a result
        self.assertEqual(len(res), n)
        for p in res:
            self.assertIn(p[Experiment.PARAMETERS]['a'], r)
            self.assertEqual(p[Experiment.PARAMETERS]['b'], 0)

        # check that each result corresponds to its parameter
        for p in res:
            self.assertEqual(p[Experiment.PARAMETERS]['a'] + p[Experiment.PARAMETERS]['b'], p[Experiment.RESULTS]['total'])

    def testStringsNotUnpacked( self ):
        '''Test we don't unpack strings, even though they're iterable'''
        v = 'A string'

        self._lab['s'] = v
        self._lab.runExperiment(NullExperiment())
        res = self._lab.results()
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0][Experiment.RESULTS]['result'], v)


if __name__ == '__main__':
    unittest.main()
